# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
.. versionadded:: 3.2

A Bootstrap "Nav" is a navigation menu: a list of items that are
links to other Views. Navs can also have a second level so that if you
click on an item, a dropdown list of vertically stacked options
appear.

This module contains the necessary Widgets and Layouts to create and
style Bootstrap Navs.


"""
from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.component.modelinterface import exposed, Field
from reahl.web.fw import Layout, Bookmark
from reahl.web.ui import AccessRightAttributes, ActiveStateAttributes, Menu, MenuItem
from reahl.web.bootstrap.ui import Div, Span, A

from reahl.component.exceptions import ProgrammerError



class Nav(Menu):
    """A Nav is a navigation menu, with items a user can click on to transition to 
    possibly different Views. Individual Items visually indicate whether they
    are active or disabled.

    .. note:: Don't be confused

       This, the :class:`reahl.web.bootstrap.navs.Nav` is not the same thing as a simple
       HTML-level :class:`reahl.web.bootstrap.ui.Nav`!

    :param view: (See :class:`~reahl.web.fw.Widget`)
    :keyword a_list: A list of :class:`~reahl.web.bootstrap.ui.A`\s representing the items.
    :keyword css_id: (See :class:`~reahl.web.ui.HTMLElement`)
    """
    add_reahl_styling = False

    def __init__(self, view, a_list=None, css_id=None):
        self.open_item = None
        super(Nav, self).__init__(view, a_list=None, css_id=None)

    @exposed
    def query_fields(self, fields):
        fields.open_item = Field(required=False, default=None)

    def create_html_representation(self):
        ul = super(Nav, self).create_html_representation()
        ul.append_class('nav')
        return ul
    
    def add_dropdown(self, title, dropdown_menu, drop_up=False, query_arguments={}):
        """Adds the dropdown_menu :class:`DropdownMenu` to this Nav. It appears as the
        top-level item with text `title`.

        :keyword drop_up: If True, the dropdown will drop upwards from its item, instead of down.
        :keyword query_arguments: (For internal use)
        """
        return self.add_submenu(title, dropdown_menu, extra_query_arguments=query_arguments, drop_up=drop_up)

    def add_html_for_item(self, item):
        li = super(Nav, self).add_html_for_item(item)
        li.append_class('nav-item')
        item.a.append_class('nav-link')
        item.a.add_attribute_source(ActiveStateAttributes(item, active_class='active'))
        item.a.add_attribute_source(AccessRightAttributes(item.a, disabled_class='disabled'))
        return li

    def add_submenu(self, title, menu, extra_query_arguments={}, **kwargs):
        if self.open_item == title:
            query_arguments={'open_item': ''}
        else:
            query_arguments={'open_item': title}
        query_arguments.update(extra_query_arguments)
        submenu = super(Nav, self).add_submenu(title, menu, query_arguments=query_arguments, **kwargs)
        return submenu
    
    def add_html_for_submenu(self, submenu, drop_up=False):
        li = super(Nav, self).add_html_for_submenu(submenu)
        if self.open_item == submenu.title:
            li.append_class('open')
        submenu.a.append_class('dropdown-toggle')
        submenu.a.set_attribute('data-toggle', 'dropdown')
        submenu.a.set_attribute('data-target', '-')
        submenu.a.add_child(Span(self.view)).append_class('caret')
        li.append_class('drop%s' % ('up' if drop_up else 'down'))
        return li



class NavLayout(Layout):
    def __init__(self, key=None, justified=False):
        super(NavLayout, self).__init__()
        self.justified = justified
        self.key = key

    @property
    def additional_css_class(self):
        return 'nav-%ss' % self.key

    def customise_widget(self):
        super(NavLayout, self).customise_widget()
        if self.key:
            self.widget.append_class(self.additional_css_class)
        if self.justified:
            self.widget.append_class('nav-justified')


class PillLayout(NavLayout):
    """This Layout makes a Nav appear as horizontally or vertically arranged pills (buttons).

    :keyword stacked: If True, the pills are stacked vertically.
    :keyword justified: If True, the pills are widened to fill all available space.
    """
    def __init__(self, stacked=False, justified=False):
        super(PillLayout, self).__init__(key='pill', justified=justified)
        if all([stacked, justified]):
            raise ProgrammerError('Pills must be stacked or justified, but not both')
        self.stacked = stacked
        self.justified = justified

    def customise_widget(self):
        super(PillLayout, self).customise_widget()
        if self.stacked:
            self.widget.append_class('nav-stacked')


class TabLayout(NavLayout):
    """This Layout makes a Nav appear as horizontal tabs.

    :keyword justified: If True, the tabs are widened to fill all available space.
    """
    def __init__(self, justified=False):
        super(TabLayout, self).__init__(key='tab', justified=justified)




class DropdownMenu(Menu):
    """A second-level menu that can be added to a Nav."""
    add_reahl_styling = False
    def create_html_representation(self):
        div = self.add_child(Div(self.view))
        self.set_html_representation(div)
        div.append_class('dropdown-menu')
        return div

    def add_html_for_item(self, item):
        self.html_representation.add_child(item.a)
        item.a.append_class('dropdown-item')
        item.a.add_attribute_source(ActiveStateAttributes(item, active_class='active'))
        item.a.add_attribute_source(AccessRightAttributes(item.a, disabled_class='disabled'))
        return item.a

    def add_submenu(self, menu, title):
        raise ProgrammerError('You cannot add a submenu to a %s.' % self.widget.__class__)



class DropdownMenuLayout(Layout):
    """Changes a DropdownMenu alignment.

    :keyword align_right: If True, align the dropdown to the right side of its parent item, else to the left.
    """
    def __init__(self, align_right=False):
        super(DropdownMenuLayout, self).__init__()
        self.align_right = align_right

    def customise_widget(self):
        if self.align_right:
            self.widget.append_class('dropdown-menu-right')




# Copyright 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import test, vassert, expected, scenario

from reahl.webdev.tools import XPath
from reahl.component.modelinterface import Field, IntegerField, exposed

from reahl.web.bootstrap.tables import Table, TableLayout, StaticColumn, TableLayout, DataTable
from reahl.web.bootstrap.ui import Div
from reahl.web_dev.fixtures import WebFixture

import reahl.web_dev.widgets.tabletests




class LayoutScenarios(WebFixture):
    @scenario
    def header_inverse(self):
        self.layout_kwargs = dict(inverse=True)
        self.expected_css_class = 'table-inverse'

    @scenario
    def border(self):
        self.layout_kwargs = dict(border=True)
        self.expected_css_class = 'table-bordered'

    @scenario
    def striped_rows(self):
        self.layout_kwargs = dict(striped=True)
        self.expected_css_class = 'table-striped'

    @scenario
    def hovered_rows(self):
        self.layout_kwargs = dict(highlight_hovered=True)
        self.expected_css_class = 'table-hover'

    @scenario
    def compacted_cells(self):
        self.layout_kwargs = dict(compact=True)
        self.expected_css_class = 'table-sm'

    @scenario
    def transposed(self):
        self.layout_kwargs = dict(transposed=True)
        self.expected_css_class = 'table-reflow'

    @scenario
    def transposed(self):
        self.layout_kwargs = dict(responsive=True)
        self.expected_css_class = 'table-responsive'


@test(LayoutScenarios)
def table_layout_options(fixture):
    """TableLayout uses Bootstrap to implement many table layout options."""

    for switched_on in [True, False]:
        layout = TableLayout(**fixture.layout_kwargs)
        Table(fixture.view).use_layout(layout)
        vassert( layout.widget.get_attribute('class') == 'table %s' % fixture.expected_css_class )



class DataItem(reahl.web_dev.widgets.tabletests.DataItem):
    @exposed
    def fields(self, fields):
        fields.row = IntegerField(label='Row', required=True, default=self.row)
        fields.alpha = Field(label='Alpha', required=True, default=self.alpha)


class DataTableFixture(reahl.web_dev.widgets.tabletests.TableFixture):

    items_per_page = 3
        
    def new_columns(self):
        return [StaticColumn(IntegerField(label='Row Number'), 'row', sort_key=lambda i: i.row),
                StaticColumn(Field(label='Alpha'), 'alpha', sort_key=lambda i: i.alpha),
                ]

    def new_data(self):
        alphas = 'THEQUICKBROWNFXJMPSVLAZYDG'
        return [DataItem(number, alpha) for number, alpha in enumerate(alphas, start=1)]
        #[(1, 'T'), (2, 'H'), (3, 'E'), (4, 'Q'),...(23, 'Z'), (24, 'Y'), (25, 'D'), (26, 'G')]

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Div):
            def __init__(self, view):
                super(MainWidget, self).__init__(view)

                data_table = DataTable(view, fixture.columns, fixture.data,
                                       items_per_page=fixture.items_per_page,
                                       caption_text='All my friends',
                                       summary='Summary for screen reader',
                                       css_id='my_table_data')
                self.add_child(data_table)
        return MainWidget

    def new_wsgi_app(self):
        return super(DataTableFixture, self).new_wsgi_app(enable_js=False,
                                                      child_factory=self.MainWidget.factory())

    def new_webconfig(self):
        webconfig = super(DataTableFixture, self).new_webconfig()
        webconfig.frontend_libraries.enable_experimental_bootstrap()
        return webconfig

    def xpath_for_sort_link_for_column(self, column_number):
        return '(//table/thead/tr/th)[%s]/a' % column_number

    def does_column_have_sort_link(self, column_number):
        column_header = self.driver_browser.web_driver.find_element_by_xpath('(//table/thead/tr/th)[%s]' % (column_number))
        return len(column_header.find_elements_by_tag_name('a')) == 1
        
    def is_column_sorted(self, column_number, direction):
        header_link = self.driver_browser.web_driver.find_element_by_xpath('(//table/thead/tr/th)[%s]/a' % (column_number))
        if direction:
            expected_class = 'sorted-%s' % direction
            return expected_class in header_link.get_attribute('class').split(' ')
        else:
            return all([expected_class not in header_link.get_attribute('class').split(' ') 
                        for expected_class in ['sorted-ascending','sorted-descending']])



@test(DataTableFixture)
def paging_through_data(fixture):
    """DataTable splits its items into different pages (between which a user can navigate), showing only the items of a particular page at a time."""
    fixture.reahl_server.set_app(fixture.wsgi_app)
    fixture.driver_browser.open('/')

    #click to last page
    fixture.driver_browser.click(XPath.link_with_text('→'))
    fixture.driver_browser.click(XPath.link_with_text('9'))
    
    vassert( fixture.table_number_rows() == 2 )
    vassert( fixture.get_table_row(1) == ['25' ,'D'] )
    vassert( fixture.get_table_row(2) == ['26' ,'G'] )

    #click to page 4
    fixture.driver_browser.click(XPath.link_with_text('←'))
    fixture.driver_browser.click(XPath.link_with_text('4'))
    
    vassert( fixture.table_number_rows() == 3 )
    vassert( fixture.get_table_row(1) == ['10' ,'R'] )
    vassert( fixture.get_table_row(2) == ['11' ,'O'] )
    vassert( fixture.get_table_row(3) == ['12' ,'W'] )


@test(DataTableFixture)
def sorting(fixture):
    """By clicking on special links in the column header, the table is sorted according to that column - ascending or descending."""
    fixture.reahl_server.set_app(fixture.wsgi_app)
    fixture.driver_browser.open('/')

    #----- by default, not sorted
    vassert( not fixture.is_column_sorted(1, 'ascending') )
    vassert( not fixture.is_column_sorted(1, 'descending') )
    vassert( not fixture.is_column_sorted(2, 'ascending') )
    vassert( not fixture.is_column_sorted(2, 'descending') )

    #----- first click on column sorts ascending
    fixture.driver_browser.click(fixture.xpath_for_sort_link_for_column(1))
    vassert( fixture.is_column_sorted(1, 'ascending') )
    vassert( fixture.get_table_row(1) == ['1' ,'T'] )
    vassert( fixture.get_table_row(2) == ['2' ,'H'] )
    vassert( fixture.get_table_row(3) == ['3' ,'E'] )
    
    #----- sort ascending on alpha, the second column
    vassert( fixture.is_column_sorted(2, None) )
    fixture.driver_browser.click(fixture.xpath_for_sort_link_for_column(2))
    vassert( fixture.is_column_sorted(2, 'ascending') )
    vassert( fixture.is_column_sorted(1, None) )
    
    vassert( fixture.get_table_row(1) == ['22' ,'A'] )
    vassert( fixture.get_table_row(2) == ['9' ,'B'] )
    vassert( fixture.get_table_row(3) == ['7' ,'C'] )

    #----- sort descending on alpha, the second column
    fixture.driver_browser.click(fixture.xpath_for_sort_link_for_column(2))
    vassert( fixture.is_column_sorted(2, 'descending') )
    
    vassert( fixture.get_table_row(1) == ['23' ,'Z'] )
    vassert( fixture.get_table_row(2) == ['24' ,'Y'] )
    vassert( fixture.get_table_row(3) == ['15' ,'X'] )

    #----- sort order stays changed when paging
    fixture.driver_browser.click(XPath.link_with_text('4'))
    
    vassert( fixture.get_table_row(1) == ['4' ,'Q'] )
    vassert( fixture.get_table_row(2) == ['18' ,'P'] )
    vassert( fixture.get_table_row(3) == ['11' ,'O'] )

    #----- contents of the page you are on changes according to a new sort order
    fixture.driver_browser.click(fixture.xpath_for_sort_link_for_column(1))
    
    vassert( fixture.get_table_row(1) == ['10' ,'R'] )
    vassert( fixture.get_table_row(2) == ['11' ,'O'] )
    vassert( fixture.get_table_row(3) == ['12' ,'W'] )


@test(DataTableFixture)
def which_columns_can_cause_sorting(fixture):
    """Only columns with sort_key specified are sortable."""

    fixture.columns.append(StaticColumn(Field(label='Not sortable'), 'alpha'))

    fixture.reahl_server.set_app(fixture.wsgi_app)
    fixture.driver_browser.open('/')

    vassert( fixture.does_column_have_sort_link(1) )
    vassert( fixture.does_column_have_sort_link(2) )
    vassert( not fixture.does_column_have_sort_link(3) )


@test(DataTableFixture)
def layout_for_contained_table(fixture):
    """You can specify a Layout to use for the actual table inside the DataTable."""

    layout = TableLayout()
    data_table = DataTable(fixture.view, fixture.columns, fixture.data, 'my_css_id', table_layout=layout)

    vassert( data_table.table.layout is layout )




# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

""".. versionadded:: 3.2

This module contains tools for controlling how elements are positioned.

The main tool here is an invisible layout grid which changes depending
on the size of the device being used to browse. The module also contains
several other related tools.

When creating a visual layout it is often useful to arrange elements
on an invisible grid. The tools here allows building such a grid, but
with a twist: the grid can resize depending on the size of the device
that is used to view it.

User devices come in a wide range of sizes. In order to be able
to change a layout depending on the size of a device, devices are classified
into several device classes: 'xs' (extra small), 'sm' (small), 'md'(medium),
'lg' (large), 'xl' (extra large).

Whenever you need to specify a size for Bootstrap grid element, you 
can specify the size that element should have *for each class of device*.

Bootstrap's grid system works on units of 1/12th the size of a given
parent width. A size for a particular device class is thus an integer
denoting a size in 1/12ths of its container's width.
"""
from __future__ import print_function, unicode_literals, absolute_import, division

import reahl.web.layout
from reahl.component.exceptions import ProgrammerError
from reahl.component.exceptions import arg_checks, IsInstance
from reahl.web.fw import Layout
from reahl.web.ui import Div, HTMLAttributeValueOption


class Container(Layout):
    """A Container Layout manages the positioning of the main areas of
    a page. By default it ensures that the HTMLElement it is used with to be
    has a size that stays at a fixed width per device class and that it stays
    centered in the horizontal.
    
    Using a Container is compulsory if you want to make use of a
    ColumnLayout.

    :keyword fluid: If True, the container fills the entire available width.

    """
    def __init__(self, fluid=False):
        super(Container, self).__init__()
        self.fluid = fluid

    def customise_widget(self):
        container_class = 'container'
        if self.fluid:
            container_class = 'container-fluid'
        self.widget.append_class(container_class)


class DeviceClass(HTMLAttributeValueOption):
    device_classes = ['xs', 'sm', 'md', 'lg', 'xl']

    def __init__(self, class_label):
        super(DeviceClass, self).__init__(class_label, class_label is not None, constrain_value_to=self.device_classes)

    @property
    def class_label(self):
        return self.option_string

    @property
    def one_smaller(self):
        index_of_one_smaller_class = self.device_classes.index(self.class_label) - 1
        if index_of_one_smaller_class < 0:
            return None
        return DeviceClass(self.device_classes[index_of_one_smaller_class])

    @property
    def all_smaller(self):
        return [DeviceClass(i) for i in self.device_classes[:self.device_classes.index(self.class_label)]]

    @classmethod
    def all_classes(cls):
        return [DeviceClass(i) for i in DeviceClass.device_classes]


class ResponsiveSize(reahl.web.layout.ResponsiveSize):
    """A size used for layouts that can adapt depending on how big the user device is.

    Sizes kwargs for each device class are given as integers that denote a number of 12ths
    of the size of the container of the element being sized. Eg: 6 would mean 6 12ths, or 
    1/2 the size of the container.

    If you specify a size for a device class, that size will be used for all devices of that
    class or bigger.

    It is not necessary to specify a size for every device class. By default, if a device
    class is omitted, it is assumed to be sized as per the nearest specified smaller device 
    class. If there is no smaller device class, a value of 12/12ths is assumed.

    :keyword xs: Size to use if the device is extra small.
    :keyword sm: Size to use if the device is small.
    :keyword md: Size to use if the device is medium.
    :keyword lg: Size to use if the device is large.
    :keyword xl: Size to use if the device is extra large.

    """
    def __init__(self, xs=None, sm=None, md=None, lg=None, xl=None):
        super(ResponsiveSize, self).__init__(xs=xs, sm=sm, md=md, lg=lg, xl=xl)
        self.offsets = {}

    def offset(self, xs=None, sm=None, md=None, lg=None, xl=None):
        self.offsets = ResponsiveSize(xs=xs, sm=sm, md=md, lg=lg, xl=xl)
        return self

    def calculated_size_for(self, device_class):
        classes_that_impact = [device_class]+device_class.all_smaller
        for possible_class in reversed(classes_that_impact):
            try:
                return self[possible_class.class_label]
            except KeyError:
                pass
        return 0

    def total_width_for(self, device_class):
        total = self.calculated_size_for(device_class)
        if self.offsets:
            total += self.offsets.calculated_size_for(device_class)
        return total

    @classmethod
    def wraps_for_some_device_class(cls, sizes):
        return any([cls.wraps_for(device_class, sizes)
                   for device_class in DeviceClass.all_classes()])

    @classmethod    
    def wraps_for(cls, device_class, sizes):
        return (cls.sum_sizes_for(device_class, sizes)) > 12

    @classmethod    
    def sum_sizes_for(cls, device_class, sizes):
        total = 0
        for size in sizes:
            total += size.total_width_for(device_class)
        return total


class ColumnLayout(reahl.web.layout.ColumnLayout):
    """A Layout that divides an element into a number of columns.

    Each argument passed to the constructor defines a column. Columns
    are added to the element using this Layout in the order they
    are passed to the constructor. Columns can also be added to the
    Widget later, by calling :meth:`ColumnLayout.add_column`.

    To define a column with a given :class:`ResponsiveSize`, pass a tuple of which
    the first element is the column name, and the second an
    instance of :class:`ResponsiveSize`.

    If an element is divided into a number of columns whose current 
    combined width is wider than 12/12ths, the overrun flows to make
    an additional row.

    It is customary, for example to specify smaller sizes (ito 12ths)
    for bigger devices where you want the columns to fit in next to each
    other, but use BIGGER sizes (such as 12/12ths) for the columns for
    smaller sized devices. This has the effect that what was displayed
    as columns next to each other on the bigger device is displayed
    as "stacked" cells on a smaller device.

    By default, the smallest device classes are sized 12/12ths.
    """
    def __init__(self, *column_definitions):
        if not all([isinstance(column_definition, tuple) for column_definition in column_definitions]):
            raise ProgrammerError('All column definitions are expected a tuple of the form (name, %s), got %s' %\
                                  (ResponsiveSize, column_definitions))
        self.added_sizes = []
        super(ColumnLayout, self).__init__(*column_definitions)

    def customise_widget(self):
        super(ColumnLayout, self).customise_widget()
        self.widget.append_class('row')

    def add_clearfix(self, column_size):
        clearfix = self.widget.add_child(Div(self.view))
        clearfix.append_class('clearfix')
        for device_class in DeviceClass.all_classes():
            if ResponsiveSize.wraps_for(device_class, self.added_sizes+[column_size]):
                clearfix.append_class('visible-%s-block' % device_class.class_label)

    def add_column(self, column_size):
        """Called to add a column of given size.

        :param column_size: A :class:`ResponsiveSize`, used to size the column.
        """
        if ResponsiveSize.wraps_for_some_device_class(self.added_sizes+[column_size]):
            self.add_clearfix(column_size)
            
        column = super(ColumnLayout, self).add_column(column_size)
        
        for device_class, value in column_size.items():
            column.append_class('col-%s-%s' % (device_class, value))
        for device_class, value in column_size.offsets.items():
            column.append_class('col-%s-offset-%s' % (device_class, value))

        self.added_sizes.append(column_size)
        return column







# Copyright 2013-2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division

import warnings
import re

from nose.tools import istest
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert, expected

from reahl.web.ui import *
from reahl.component.modelinterface import BooleanField
from reahl.web.fw import Bookmark, Url
from reahl.webdev.tools import WidgetTester, XPath, Browser
from reahl.web_dev.fixtures import WebFixture

_ = Translator('reahl-web')

@istest
class BasicReahlWidgets(object):
    @test(WebFixture)
    def priority_group(self, fixture):
        """A Priority group is a construct that ensures only one of its members has primary priority."""
        g = PriorityGroup()
        w1 = P(fixture.view)
        w2 = P(fixture.view)
        w3 = P(fixture.view)
        
        g.add_secondary(w1)
        g.add_primary(w2)
        g.add_secondary(w3)

        vassert( w1.attributes.v['class'] == 'reahl-priority-secondary' )
        vassert( w2.attributes.v['class'] == 'reahl-priority-primary' )
        vassert( w3.attributes.v['class'] == 'reahl-priority-secondary' )
        
        # You cannot have more than one primary
        w4 = P(fixture.view)
        with expected(AssertionError):
            g.add_primary(w4)

        # You cannot add something more than once
        with expected(AssertionError):
            g.add_secondary(w3)


    class Scenarios(WebFixture):
        @scenario
        def panel(self):
            self.widget = Div(self.view)
            self.expected_html = '<div></div>'

        @scenario
        def input_group1(self):
            self.widget = InputGroup(self.view)
            self.expected_html = '<fieldset></fieldset>'

        @scenario
        def input_group2(self):
            self.widget = InputGroup(self.view, label_text='text')
            self.expected_html = '<fieldset><label>text</label></fieldset>'


                                
    @test(Scenarios)
    def aliassed_widgets(self, fixture):
        """Some Widgets really also just render HTML, but have more abstract names."""
        tester = WidgetTester(fixture.widget)
        rendered_html = tester.render_html()
        vassert( rendered_html == fixture.expected_html )

    @test(WebFixture)
    def html5_page(self, fixture):
        """An HTML5Page is an empty HTML 5 document using the header and body widgets."""
        widget = HTML5Page(fixture.view, title='It: $current_title')
        widget.add_default_slot('slot1', P.factory())
        tester = WidgetTester(widget)
        
        rendered_html = tester.render_html()
        head = '<head><title>It: %s</title></head>' % fixture.view.title
        expected_regex = '<!DOCTYPE html><html class="no-js"><script>.*</script>%s<body></body></html>' % head
        vassert( re.match(expected_regex, rendered_html.replace('\n', '')) )
        
        vassert( list(widget.default_slot_definitions.keys()) == ['slot1'] )

    @test(WebFixture)
    def rendering_a_menu_item(self, fixture):
        """A MenuItem is a li with a in it."""
        
        description = 'The link'
        href_path = '/link'
        bookmark = Bookmark('/', href_path, description)

        menu = Menu(fixture.view)
        menu_item = MenuItem.from_bookmark(fixture.view, bookmark)
        menu.add_item(menu_item)
        tester = WidgetTester(menu)

        with fixture.context:
            actual = tester.get_html_for('//li')
            
        vassert( actual == '<li><a href="%s">%s</a></li>' % (href_path, description) )
        
    @test(WebFixture)
    def menu(self, fixture):
        """Menus can be populated with a list of A's or Bookmarks, or MenuItems."""
        # Case: a normal menu from bookmarks
        item_specs = [Bookmark('/', '/href1', 'description1'),
                      Bookmark('/', '/href2', 'description2')]
        menu = Menu(fixture.view).with_bookmarks(item_specs)
        tester = WidgetTester(menu)

        with fixture.context:
            actual = tester.render_html()
        rendered_children = WidgetTester(menu.children[0].children).render_html()
        expected_html = '''<ul class="reahl-menu">%s</ul>''' % rendered_children
        vassert( actual == expected_html )
        
        #case: using A's
        a_list = [A.from_bookmark(fixture.view, i) for i in item_specs]
        menu = Menu(fixture.view).with_a_list(a_list)
        tester = WidgetTester(menu)
        with fixture.context:
            actual = tester.render_html()
        vassert( actual == expected_html )
        
        # Case: adding already constructed menu item
        menu = Menu(fixture.view)
        item1 = MenuItem(fixture.view, A.from_bookmark(fixture.view, item_specs[0]))
        item2 = MenuItem(fixture.view, A.from_bookmark(fixture.view, item_specs[1]))
        menu.add_item(item1)
        menu.add_item(item2)
        vassert( menu.menu_items == [item1, item2] )
        
        tester = WidgetTester(menu)
        with fixture.context:
            actual = tester.render_html()
        vassert( actual == expected_html )
        
        # Case: Menu
        menu = Menu(fixture.view).with_bookmarks(item_specs)
        vassert( menu.html_representation.attributes.v['class'] == 'reahl-menu' )

        # Case: HMenu
        with warnings.catch_warnings(record=True):
            menu = HMenu(fixture.view, []).with_bookmarks(item_specs)
        vassert( menu.html_representation.attributes.v['class'] == 'reahl-horizontal reahl-menu' )

        # Case: VMenu
        with warnings.catch_warnings(record=True):
            menu = VMenu(fixture.view, []).with_bookmarks(item_specs)
        vassert( menu.html_representation.attributes.v['class'] == 'reahl-menu reahl-vertical' )

    @test(WebFixture)
    def menu_can_have_submenus(self, fixture):
        """One Menu can be added to another as submenu."""
        # Case: a normal menu from bookmarks
        single_item_bookmark = Bookmark('/', '/href3', 'description3')
        item_specs = [Bookmark('/', '/href1', 'description1'),
                      Bookmark('/', '/href2', 'description2')]

        sub_menu = Menu(fixture.view).with_bookmarks(item_specs)
        sub_menu_title = 'Subbie'
        menu = Menu(fixture.view)
        menu.add_item(MenuItem(fixture.view, A.from_bookmark(fixture.view, single_item_bookmark)))
        menu.add_submenu(sub_menu_title, sub_menu)

        expected_html = '''<ul class="reahl-menu">'''\
                   '''<li><a href="/href3">description3</a></li>'''\
                   '''<li><a>Subbie</a>'''\
                   '''<ul class="reahl-menu">'''\
                   '''<li><a href="/href1">description1</a></li>'''\
                   '''<li><a href="/href2">description2</a></li>'''\
                   '''</ul>'''\
                   '''</li>'''\
                   '''</ul>'''
        tester = WidgetTester(menu)
        with fixture.context:
            actual = tester.render_html()
        vassert( actual == expected_html )

    @test(WebFixture)
    def menu_can_have_submenus_deprecated_interface(self, fixture):
        """Menus can have sub-menus too - when built-up using .add_item()."""
        # Case: a normal menu from bookmarks
        single_item_bookmark = Bookmark('/', '/href3', 'description3')
        item_specs = [Bookmark('/', '/href1', 'description1'),
                      Bookmark('/', '/href2', 'description2')]

        sub_menu = Menu(fixture.view).with_bookmarks(item_specs)
        sub_menu_title = 'Subbie'
        menu = Menu(fixture.view)
        menu.add_item(MenuItem(fixture.view, A.from_bookmark(fixture.view, single_item_bookmark)))
        menu.add_submenu(sub_menu_title, sub_menu)

        expected_html = '''<ul class="reahl-menu">'''\
                   '''<li><a href="/href3">description3</a></li>'''\
                   '''<li><a>Subbie</a>'''\
                   '''<ul class="reahl-menu">'''\
                   '''<li><a href="/href1">description1</a></li>'''\
                   '''<li><a href="/href2">description2</a></li>'''\
                   '''</ul>'''\
                   '''</li>'''\
                   '''</ul>'''
        tester = WidgetTester(menu)
        with fixture.context:
            actual = tester.render_html()
        vassert( actual == expected_html )

    class MenuItemScenarios(WebFixture):
        description = 'The link'
        href = Url('/link')

        @scenario
        def not_active(self):
            self.active_regex = None
            self.request.environ['PATH_INFO'] = '/something/else'
            self.active = False

        @scenario
        def active_exact_path(self):
            self.active_regex = None
            self.request.environ['PATH_INFO'] = '/link'
            self.active = True

        @scenario
        def active_partial_path(self):
            self.active_regex = None
            self.request.environ['PATH_INFO'] = '/link/something/more'
            self.active = True

        @scenario
        def inactive_partial_path(self):
            self.active_regex = '^/link$'
            self.request.environ['PATH_INFO'] = '/link/something/more'
            self.active = False

    @test(MenuItemScenarios)
    def rendering_active_menu_items(self, fixture):    
        """A MenuItem is marked as active based on its active_regex or the A it represents."""
        description = 'The link'
        href = Url('/link')

        menu = Menu(fixture.view)
        menu_item = MenuItem(fixture.view, A(fixture.view, href, description=description), active_regex=fixture.active_regex)
        menu.add_item(menu_item)
        tester = WidgetTester(menu)

        actual = tester.get_html_for('//li')
        class_str = '' if not fixture.active else ' class="active"'
        expected_menu_item_html = '<li%s><a href="/link">The link</a></li>' % (class_str)

        vassert( actual == expected_menu_item_html )


    class CustomMenuItemFixture(WebFixture):
        def new_href(self):
            return Url('/link')

        def new_menu_item(self):
            description = 'The link'
            href = Url('/link')
        
            menu_item = MenuItem(self.view, A(self.view, self.href, description=description))
            return menu_item

        def new_menu(self):
            menu = Menu(self.view)
            menu.add_item(self.menu_item)
            return menu
        
        def new_tester(self):
            return WidgetTester(self.menu)

        def item_displays_as_active(self):
            actual = self.tester.get_html_for('//li')
            class_str = ' class="active"'
            expected_menu_item_html = '<li%s><a href="/link">The link</a></li>' % (class_str)
            return actual == expected_menu_item_html

        def set_request_url(self, href):
            self.request.environ['PATH_INFO'] = str(href)
    
    @test(CustomMenuItemFixture)
    def custom_active_menu_items(self, fixture):
        """You can specify a custom method by which a MenuItem determines its active state."""

        # The default behaviour happens when no custom method is supplied
        fixture.set_request_url(fixture.href)
        vassert( fixture.item_displays_as_active() )
        
        # Overriding behaviour happens when supplied
        fixture.menu_item.determine_is_active_using(lambda: False)
        vassert( not fixture.item_displays_as_active() )

        url_on_which_item_is_usually_inactive = Url('/another_href')
        fixture.set_request_url(url_on_which_item_is_usually_inactive)
        fixture.menu_item.determine_is_active_using(lambda: True)
        vassert( fixture.item_displays_as_active() )

    @test(WebFixture)
    def language_menu(self, fixture):
        """A Menu can also be constructed to let a user choose to view the same page in 
           another of the supported languages."""
        
        class PanelWithMenu(Div):
            def __init__(self, view):
                super(PanelWithMenu, self).__init__(view)
                self.add_child(Menu(view).with_languages())
                self.add_child(P(view, text=_('This is an English sentence.')))

        wsgi_app = fixture.new_wsgi_app(child_factory=PanelWithMenu.factory())
        
        browser = Browser(wsgi_app)
        browser.open('/')
        
        vassert( browser.is_element_present(XPath.paragraph_containing('This is an English sentence.')) )

        browser.click(XPath.link_with_text('Afrikaans'))
        vassert( browser.is_element_present(XPath.paragraph_containing('Hierdie is \'n sin in Afrikaans.')) )

        browser.click(XPath.link_with_text('English (United Kingdom)'))
        vassert( browser.is_element_present(XPath.paragraph_containing('This is an English sentence.')) )


    class Scenarios(WebFixture):
        @scenario
        def yuidoc(self):
            self.config.web.frontend_libraries.use_deprecated_yui()
            with warnings.catch_warnings(record=True):
                self.widget = YuiDoc(self.view, 'docid', 'docclass')
            self.expected_html = '<div id="docid" class="docclass"><div id="hd" class="yui-g"><header></header></div><div id="bd" role="main"><div id="yui-main"><div class="yui-b"></div></div><div class="yui-b"></div></div><div id="ft"><footer></footer></div></div>'
        @scenario
        def yuiblock(self):
            self.config.web.frontend_libraries.use_deprecated_yui()
            with warnings.catch_warnings(record=True):
                self.widget = YuiBlock(self.view)
            self.expected_html = '<div class="yui-b"></div>'

        @scenario
        def yuigrid(self):
            self.config.web.frontend_libraries.use_deprecated_yui()
            with warnings.catch_warnings(record=True):
                self.widget = YuiGrid(self.view)
            self.expected_html = '<div class="yui-g"></div>'

        @scenario
        def yuiunit1(self):
            self.config.web.frontend_libraries.use_deprecated_yui()
            with warnings.catch_warnings(record=True):
                self.widget = YuiUnit(self.view)
            self.expected_html = '<div class="yui-u"></div>'

        @scenario
        def yuiunit_first(self):
            self.config.web.frontend_libraries.use_deprecated_yui()
            with warnings.catch_warnings(record=True):
                self.widget = YuiUnit(self.view, first=True)
            self.expected_html = '<div class="first yui-u"></div>'
            
    @test(Scenarios)
    def yui_widgets(self, fixture):
        """Different Yui widgets for rendering HTML that it compatible with the Yui CSS framework."""
        tester = WidgetTester(fixture.widget)
        rendered_html = tester.render_html()
        vassert( rendered_html == fixture.expected_html )
        
    @test(WebFixture)
    def twocolumn_page(self, fixture):
        """A simple Yui page with two columns, a header and a footer."""
        fixture.config.web.frontend_libraries.use_deprecated_yui()
        with warnings.catch_warnings(record=True):
            widget = TwoColumnPage(fixture.view, title='It: $current_title')
        widget.add_default_slot('slot1', P.factory())
        tester = WidgetTester(widget)
        
        rendered_html = tester.render_html()
        expected_regex = '<!DOCTYPE html><html class="no-js"><script>.*</script><head><title>It: A view</title></head><body><div id="doc" class="yui-t2"><div id="hd" class="yui-g"><header></header></div><div id="bd" role="main"><div id="yui-main"><div class="yui-b"></div></div><div class="yui-b"></div></div><div id="ft"><footer></footer></div></div></body></html>'
        vassert( re.match(expected_regex, rendered_html.replace('\n','') ))
        
        vassert( list(widget.default_slot_definitions.keys()) == ['slot1'] )
        
        vassert( isinstance(widget.yui_page, YuiDoc) )
        vassert( widget.footer is widget.yui_page.footer )
        vassert( widget.header is widget.yui_page.header )
        vassert( widget.main is widget.yui_page.main_block )
        vassert( widget.secondary is widget.yui_page.secondary_block )




class TabbedPanelAjaxFixture(WebFixture):
    def new_PopulatedTabbedPanel(self):
        fixture = self
        class PopulatedTabbedPanel(TabbedPanel):
            def __init__(self, view):
                super(PopulatedTabbedPanel, self).__init__(view, 'tabbed')
                multi_tab = MultiTab(view, 'multitab name', 'multi-main', P.factory(text='main multi tab content'))
                tab1 = Tab(fixture.view, 'tab 1 name', 'multi1', P.factory(text='tab 1/1 content'))
                tab2 = Tab(fixture.view, 'tab 2 name', 'multi2', P.factory(text='tab 1/2 content'))
                multi_tab.add_tab(tab1)
                multi_tab.add_tab(tab2)
                self.add_tab(multi_tab)

                tab3 = Tab(fixture.view, 'tab 3 name', 'tab3', P.factory(text='tab 3 content'))
                self.add_tab(tab3)
        return PopulatedTabbedPanel

    def new_wsgi_app(self):
        return super(TabbedPanelAjaxFixture, self).new_wsgi_app(enable_js=True,
                                                   child_factory=self.PopulatedTabbedPanel.factory())
    def tab_is_active(self, tab_name):
        return self.driver_browser.execute_script('return window.jQuery("a:contains(\'%s\')").parent().hasClass("active")' % tab_name)

    def tab_contents_equals(self, expected_contents):
        return self.driver_browser.execute_script('return window.jQuery("div.reahl-tabbedpanel div").html() == "%s"' % expected_contents)


@istest
class TabbedPanelTests(object):
    @test(WebFixture)
    def basic_rendering(self, fixture):
        """A TabbedPanel is a Div which contains a Horizontal Menu and a Div."""
        fixture.request.query_string = 'tab=tab1'
        tabbed_panel = TabbedPanel(fixture.view, 'tabbed_name')
        tabbed_panel.add_tab(Tab(fixture.view, 'tab 1 name', 'tab1', P.factory(text='tab 1 content')))

        tester = WidgetTester(tabbed_panel)

        expected_html = '''<div id="tabbed_name" class="reahl-tabbedpanel">'''\
                        '''<ul class="reahl-horizontal reahl-menu">'''\
                        '''<li class="active"><a href="/?tab=tab1" class="reahl-ajaxlink">tab 1 name</a></li>'''\
                        '''</ul>'''\
                        '''<div><p>tab 1 content</p></div>'''\
                        '''</div>'''
        actual = tester.render_html()        
        vassert( actual == expected_html )

    @test(WebFixture)
    def tabs_with_sub_options(self, fixture):
        """A TabbedPanel can have Tabs that are each composed of multiple sub-options."""
        fixture.request.query_string = 'tab=mult2'
        tabbed_panel = TabbedPanel(fixture.view, 'tabbed_name')
        multi_tab = MultiTab(fixture.view, 'tab 1 name', 'multitab-main', P.factory(text='multi tab main content'))
        multi_tab.add_tab(Tab(fixture.view, 'multi tab 1', 'mult1', P.factory(text='tab 1/1 content')))
        multi_tab.add_tab(Tab(fixture.view, 'multi tab 2', 'mult2', P.factory(text='tab 1/2 content')))
        tabbed_panel.add_tab(multi_tab)

        tester = WidgetTester(tabbed_panel)
        
        expected_html = '''<div id="tabbed_name" class="reahl-tabbedpanel">'''\
                        '''<ul class="reahl-horizontal reahl-menu">'''\
                        '''<li class="active"><a href="/?tab=multitab-main" class="reahl-ajaxlink">tab 1 name</a>&nbsp;'''\
                        '''<a class="dropdown-handle">▼</a>'''\
                        '''<ul class="reahl-menu reahl-vertical">'''\
                        '''<li><a href="/?tab=mult1" class="reahl-ajaxlink">multi tab 1</a></li>'''\
                        '''<li class="active"><a href="/?tab=mult2" class="reahl-ajaxlink">multi tab 2</a></li>'''\
                        '''</ul>'''\
                        '''</li>'''\
                        '''</ul>'''\
                        '''<div><p>tab 1/2 content</p></div>'''\
                        '''</div>'''
        actual = tester.render_html()        
        vassert( actual == expected_html )



    class DefaultTabScenarios(WebFixture):
        @scenario
        def specified_on_query_string(self):
            self.request.query_string = 'tab=tab2'
            self.expected_contents = '<p>tab 2 content</p>'
            self.tab1_active = False
            self.tab2_active = True
            
        @scenario
        def defaulted(self):
            self.request.query_string = ''
            self.expected_contents = '<p>tab 1 content</p>'
            self.tab1_active = True
            self.tab2_active = False
            
    @test(DefaultTabScenarios)
    def default_active_tab(self, fixture):
        """The first tab is active by default (if the active tab is not indicated in the query_string)."""
        tab1 = Tab(fixture.view, 'tab 1 name', 'tab1', P.factory(text='tab 1 content'))
        tab2 = Tab(fixture.view, 'tab 2 name', 'tab2', P.factory(text='tab 2 content'))

        tabbed_panel = TabbedPanel(fixture.view, 'tabbed_name')
        tabbed_panel.add_tab(tab1)
        tabbed_panel.add_tab(tab2)

        [menu_item1, menu_item2] = tabbed_panel.menu.menu_items
        vassert( menu_item1.is_active == fixture.tab1_active )
        vassert( menu_item2.is_active == fixture.tab2_active )

        tester = WidgetTester(tabbed_panel)
        panel_contents = tester.get_html_for('//div/div/*')
        vassert( panel_contents == fixture.expected_contents )

    class DefaultMultiTabScenarios(WebFixture):
        @scenario
        def specified_on_query_string(self):
            self.request.query_string = 'tab=multi2'
            self.expected_contents = '<p>tab 1/2 content</p>'
            self.multi_tab_active = True
            self.tab1_active = False
            self.tab2_active = True
            self.tab3_active = False
            
        @scenario
        def defaulted(self):
            self.request.query_string = ''
            self.expected_contents = '<p>multi tab main content</p>'
            self.multi_tab_active = True
            self.tab1_active = False
            self.tab2_active = False
            self.tab3_active = False

        @scenario
        def specified_to_other_tab(self):
            self.request.query_string = 'tab=tab3'
            self.expected_contents = '<p>tab 3 content</p>'
            self.multi_tab_active = False
            self.tab1_active = False
            self.tab2_active = False
            self.tab3_active = True

    @test(DefaultMultiTabScenarios)
    def default_active_multi_tab(self, fixture):
        """The first item of the first tab is active by default (if the active tab is not indicated in the query_string)."""
        tabbed_panel = TabbedPanel(fixture.view, 'tabbed_panel')
        multi_tab = MultiTab(fixture.view, 'tab 1 name', 'multi-tab', P.factory(text='multi tab main content'))
        tab1 = Tab(fixture.view, 'tab 1 name', 'multi1', P.factory(text='tab 1/1 content'))
        tab2 = Tab(fixture.view, 'tab 2 name', 'multi2', P.factory(text='tab 1/2 content'))
        multi_tab.add_tab(tab1)
        multi_tab.add_tab(tab2)
        tabbed_panel.add_tab(multi_tab)

        tab3 = Tab(fixture.view, 'tab 3 name', 'tab3', P.factory(text='tab 3 content'))
        tabbed_panel.add_tab(tab3)

        [top_level_menu_item_for_multitab, top_level_normal_menu_item] = tabbed_panel.menu.menu_items
        [multi_tab_menu_item1, multi_tab_menu_item2] = multi_tab.menu.menu_items

        vassert( top_level_menu_item_for_multitab.is_active == fixture.multi_tab_active )
        vassert( multi_tab_menu_item1.is_active == fixture.tab1_active )
        vassert( multi_tab_menu_item2.is_active == fixture.tab2_active )
        vassert( top_level_normal_menu_item.is_active == fixture.tab3_active )

        tester = WidgetTester(tabbed_panel)
        panel_contents = tester.get_html_for('//div/div/*')
        vassert( panel_contents == fixture.expected_contents )

    @test(TabbedPanelAjaxFixture)
    def switching_panels(self, fixture):
        """The contents registered with the TabbedPanel's active tab are displayed."""
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open('/')

        # by default, the first tab is active and its contents are displayed
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, 'multitab name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, '<p>main multi tab content</p>') )

        fixture.driver_browser.click(XPath.link_with_text('tab 3 name'))
        # tab3 is active and its contents are shown
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, 'tab 3 name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, '<p>tab 3 content</p>') )

        fixture.driver_browser.mouse_over(XPath.link_with_text('multitab name'))
        fixture.driver_browser.click(XPath.link_with_text('tab 2 name'))
        # tab2 is active and its contents are shown
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, 'tab 2 name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, '<p>tab 1/2 content</p>') )

        fixture.driver_browser.click(XPath.link_with_text('▼'))
        # clicking on the down arrow does nothing
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, 'tab 2 name') )

        fixture.driver_browser.click(XPath.link_with_text('multitab name'))
        # the main multitab is active and its contents are shown
        vassert( fixture.driver_browser.wait_for(fixture.tab_is_active, 'multitab name') )
        vassert( fixture.driver_browser.wait_for(fixture.tab_contents_equals, '<p>main multi tab content</p>') )



class SlidingPanelFixture(WebFixture):
    def new_PopulatedSlidingPanel(self):
        fixture = self
        class PopulatedSlidingPanel(SlidingPanel):
            def __init__(self, view):
                super(PopulatedSlidingPanel, self).__init__(view, 'slide')
                panel0 = Div(view)
                panel0.add_child(P(view, text='Contents for panel 0'))
                self.add_panel(panel0)

                panel1 = Div(view)
                panel1.add_child(P(view, text='Contents for panel 1'))
                self.add_panel(panel1)

        return PopulatedSlidingPanel

    def new_wsgi_app(self):
        return super(SlidingPanelFixture, self).new_wsgi_app(enable_js=True,
                                                           child_factory=self.PopulatedSlidingPanel.factory())
    def panel_is_visible(self, index):
        return self.driver_browser.wait_for_element_visible(XPath.paragraph_containing('Contents for panel %s' % index))

    def panel_is_not_visible(self, index):
        return self.driver_browser.wait_for_element_not_visible(XPath.paragraph_containing('Contents for panel %s' % index))



@test(SlidingPanelFixture)
def opens_on_selected_index(fixture):
    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser
    
    browser.open('/?index=1')
    vassert( fixture.panel_is_visible(1) )
    

@test(SlidingPanelFixture)
def slide_right(fixture):
    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser

    browser.open('/')
    vassert( fixture.panel_is_visible(0) )

    browser.click(XPath.link_with_text('>'))
    vassert( fixture.panel_is_not_visible(0) )
    vassert( fixture.panel_is_visible(1) )

    #wrapping back to start
    browser.click(XPath.link_with_text('>'))
    vassert( fixture.panel_is_not_visible(1) )
    vassert( fixture.panel_is_visible(0) )

@test(SlidingPanelFixture)
def slide_left(fixture):
    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser

    browser.open('/')
    vassert( fixture.panel_is_visible(0) )

    #wrapping
    browser.click(XPath.link_with_text('<'))
    vassert( fixture.panel_is_not_visible(0) )
    vassert( fixture.panel_is_visible(1) )

    browser.click(XPath.link_with_text('<'))
    vassert( fixture.panel_is_not_visible(1) )
    vassert( fixture.panel_is_visible(0) )

class PopupAFixture(WebFixture):
    # (note that this xpath ensures that the p is the ONLY content of the dialog)
    poppedup_contents = "//div[@class='reahl-dialogcontent' and count(*)=1]/p[@id='contents']"
    

@istest
class PopupATests(object):
    @test(PopupAFixture)
    def default_behaviour(self, fixture):
        """If you click on the A, a popupwindow opens with its contents the specified
           element on the target page."""

        class PopupTestPanel(Div):
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
                popup_contents = self.add_child(P(view, text='this is the content of the popup'))
                popup_contents.set_id('contents')

        wsgi_app = fixture.new_wsgi_app(child_factory=PopupTestPanel.factory(), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)
        fixture.driver_browser.open('/')
        
        # The A is rendered correctly
        fixture.driver_browser.is_element_present("//a[@title='Home page' and text()='Home page' and @href='/']")

        # subsequent behaviour
        fixture.driver_browser.click(XPath.link_with_text('Home page'))
        fixture.driver_browser.wait_for_element_visible(fixture.poppedup_contents)
        
        fixture.driver_browser.click(XPath.button_labelled('Close'))
        fixture.driver_browser.wait_for_element_not_visible(fixture.poppedup_contents)

    @test(PopupAFixture)
    def customising_dialog_buttons(self, fixture):
        """The buttons of the dialog can be customised."""
        
        class PopupTestPanel(Div):
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                popup_a = self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
                popup_a.add_button(DialogButton('Butt1'))
                popup_a.add_button(DialogButton('Butt2'))
                popup_contents = self.add_child(P(view, text='this is the content of the popup'))
                popup_contents.set_id('contents')

        wsgi_app = fixture.new_wsgi_app(child_factory=PopupTestPanel.factory(), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)

        button1_xpath = XPath.button_labelled('Butt1')
        button2_xpath = XPath.button_labelled('Butt2')
        fixture.driver_browser.open('/')

        fixture.driver_browser.click(XPath.link_with_text('Home page'))
        fixture.driver_browser.wait_for_element_visible(fixture.poppedup_contents)

        vassert( fixture.driver_browser.is_element_present(button1_xpath) )
        vassert( fixture.driver_browser.is_element_present(button2_xpath) )

    @test(PopupAFixture)
    def workings_of_check_checkbox_button(self, fixture):
        """A CheckCheckBoxButton checks the checkbox on the original page when clicked."""

        class PopupTestPanel(Div):
            @exposed
            def fields(self, fields):
                fields.field = BooleanField()
                
            def __init__(self, view):
                super(PopupTestPanel, self).__init__(view)
                popup_a = self.add_child(PopupA(view, view.as_bookmark(), '#contents'))
                popup_contents = self.add_child(P(view, text='this is the content of the popup'))
                popup_contents.set_id('contents')
                form = self.add_child(Form(view, 'aform'))
                checkbox = form.add_child(CheckboxInput(form, self.fields.field))

                popup_a.add_button(CheckCheckboxButton('Checkit', checkbox))

        wsgi_app = fixture.new_wsgi_app(child_factory=PopupTestPanel.factory(), enable_js=True)
        fixture.reahl_server.set_app(wsgi_app)

        button_xpath = XPath.button_labelled('Checkit')
        fixture.driver_browser.open('/')

        fixture.driver_browser.click(XPath.link_with_text('Home page'))
        fixture.driver_browser.wait_for_element_visible(fixture.poppedup_contents)

        fixture.driver_browser.click(button_xpath)
        fixture.driver_browser.wait_for_element_not_visible(fixture.poppedup_contents)

        vassert( fixture.driver_browser.is_checked("//input[@type='checkbox']") )







# Copyright 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six
import time
import os
import threading 

import webob


from reahl.tofu import vassert, scenario, expected, test, Fixture, temp_file_with
from reahl.stubble import stubclass


from reahl.webdev.tools import XPath
from reahl.web_dev.fixtures import WebFixture, WebBasicsMixin

from reahl.sqlalchemysupport import Session
from reahl.webdeclarative.webdeclarative import PersistedFile
from reahl.component.modelinterface import exposed, Field, Event, FileField, Action, ValidationConstraint
from reahl.component.exceptions import DomainException
from reahl.web.fw import WebExecutionContext

from reahl.web.bootstrap.ui import Span, Div, P, TextNode
from reahl.web.bootstrap.forms import Form, Label
from reahl.web.bootstrap.files import FileUploadInput, FileUploadPanel, Button, FormLayout, FileInput, FileInputButton





#----------------------------------
class FileInputButtonFixture(Fixture, WebBasicsMixin):

    def upload_button_indicates_focus(self):
        element = self.driver_browser.find_element(XPath.label_with_text('Choose file(s)'))
        return 'focus' in element.get_attribute('class')

    def new_domain_object(self):
        class DomainObject(object):
            files = []
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label='Attached files')

            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')

        return DomainObject()

    def new_FileUploadForm(self):
        fixture = self
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.add_child(FileInputButton(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))
        return FileUploadForm   

    def new_wsgi_app(self, enable_js=False):
        return super(FileInputButtonFixture, self).new_wsgi_app(child_factory=self.FileUploadForm.factory(), enable_js=enable_js)

    def new_webconfig(self):
        web = super(FileInputButtonFixture, self).new_webconfig()
        web.frontend_libraries.enable_experimental_bootstrap()
        return web


@test(FileInputButtonFixture)
def file_upload_button(fixture):
    """A FileInputButton lets you upload files using the browser's file choosing mechanism."""

    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = fixture.driver_browser
    browser.open('/')
    
    file_to_upload = temp_file_with('some content')

    browser.type(XPath.input_labelled('Choose file(s)'), file_to_upload.name)
    
    vassert( len(fixture.domain_object.files) == 0 )
    browser.click(XPath.button_labelled('Submit'))
    vassert( len(fixture.domain_object.files) == 1 )


@test(FileInputButtonFixture)
def file_upload_button_focus(fixture):
    """If the FileInputButton gets tab focus, it is styled to appear focussed."""

    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = fixture.driver_browser
    browser.open('/')

    vassert( browser.wait_for_not(fixture.upload_button_indicates_focus) )
    browser.focus_on(XPath.input_labelled('Choose file(s)'))
    vassert( browser.wait_for(fixture.upload_button_indicates_focus) )


#@test(FilesUploadFixture)
#def select_file_dialog_opens(fixture):
#    """Clicking on the FileInputButton opens up the browser choose file dialog."""
#
#    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
#    browser = fixture.driver_browser
#    browser.open('/')
#
#    browser.click(fixture.upload_button_xpath)
#    # We can't dismiss the local file dialog with selenium


class FileInputFixture(FileInputButtonFixture):
    def new_FileUploadForm(self):
        fixture = self
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.add_child(FileInput(self, fixture.domain_object.fields.files))
        return FileUploadForm   

    message_span_xpath = '//div[contains(@class, "reahl-bootstrapfileinput")]//span[2]'
    def message_displayed_is(self, message):
        return message == self.driver_browser.get_inner_html_for(self.message_span_xpath)

    def message_is_visible(self):
        return self.driver_browser.is_visible(self.message_span_xpath)

    def standard_file_input_is_visible(self):
        input_element = self.driver_browser.find_element(XPath.input_labelled('Choose file(s)'))
        return not input_element.value_of_css_property('width').startswith('0.')


@test(FileInputFixture)
def file_input_basics(fixture):
    """A FileInput is a FileInputButton combined with a area where the chosen file name is displayed."""
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = fixture.driver_browser
    browser.open('/')

    browser.wait_for(fixture.message_displayed_is, 'No files chosen')
    browser.type(XPath.input_labelled('Choose file(s)'), '/tmp/koos.html')
    browser.wait_for(fixture.message_displayed_is, 'koos.html')

    browser.type(XPath.input_labelled('Choose file(s)'), '/tmp/koos.html\n/tmp/jannie.html')
    browser.wait_for(fixture.message_displayed_is, '2 files chosen')


@test(FileInputFixture)
def i18n(fixture):
    """All messages have translations."""
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = fixture.driver_browser
    browser.open('/af/')

    browser.wait_for(fixture.message_displayed_is, 'Geen lêers gekies')
    browser.wait_for_element_present(XPath.input_labelled('Kies lêer(s)'))

    browser.type(XPath.input_labelled('Kies lêer(s)'), '/tmp/koos.html\n/tmp/jannie.html')
    browser.wait_for(fixture.message_displayed_is, '2 gekose lêers')


@test(FileInputFixture)
def file_input_without_js(fixture):
    """If JS available, we display a bootstrap-styled button, and associated span that look like a pretty version of a standard file input; otherwise we degrade to displaying only the standard file input"""
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = fixture.driver_browser
    browser.open('/')

    browser.wait_for(fixture.message_is_visible)
    browser.wait_for_not(fixture.standard_file_input_is_visible)
    
    browser.switch_styling(javascript=False)

    browser.wait_for_not(fixture.message_is_visible)
    browser.wait_for(fixture.standard_file_input_is_visible)







class FileUploadInputFixture(WebFixture):
    def file_was_uploaded(self, filename):
        return Session.query(PersistedFile).filter_by(filename=os.path.basename(filename)).count() == 1

    file_to_upload1_name = 'file1.html'
    file_to_upload2_name = 'file2.gif'
    file_to_upload1_content = b'some content'
    file_to_upload2_content = b'some different content'

    def new_file_to_upload1(self):
        return temp_file_with(self.file_to_upload1_content, name=self.file_to_upload1_name, mode='w+b')

    def new_file_to_upload2(self):
        return temp_file_with(self.file_to_upload2_content, name=self.file_to_upload2_name, mode='w+b')

    def new_domain_object(self):
        class DomainObject(object):
            def __init__(self):
                self.throws_exception = False
                self.files = []
                self.submitted_file_info = {}
                self.submitted = False
               
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label='Attached files', required=True)

            @exposed
            def events(self, events):
                events.submit = Event(label='Submit', action=Action(self.submit))
                
            def submit(self):
                if self.throws_exception:
                    raise DomainException()
                for f in self.files:
                    with f.open() as opened_file:
                        contents = opened_file.read()
                    self.submitted_file_info[f.filename] = (contents, f.mime_type)
                self.submitted = True
                
        return DomainObject()

    def new_FileUploadForm(self):
        fixture = self
        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.set_attribute('novalidate','novalidate')
                self.use_layout(FormLayout())
                self.layout.add_input(FileUploadInput(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))

        return FileUploadForm

    def new_wsgi_app(self, enable_js=False):
        return super(FileUploadInputFixture, self).new_wsgi_app(child_factory=self.FileUploadForm.factory(), enable_js=enable_js)

    def uploaded_file_is_listed(self, filename):
        return self.driver_browser.is_element_present('//ul/li/span[text()="%s"]/../input[@value="Remove"]' % os.path.basename(filename))

    def upload_file_is_queued(self, filename):
        return self.driver_browser.is_element_present('//ul/li/span[text()="%s"]/../input[@value="Cancel"]' % os.path.basename(filename))

    def new_webconfig(self):
        web = super(FileUploadInputFixture, self).new_webconfig()
        web.frontend_libraries.enable_experimental_bootstrap()
        return web


class ConstrainedFileUploadInputFixture(FileUploadInputFixture):
    def new_domain_object(self):
        fixture = self
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.files = fixture.file_field

            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')

        return DomainObject()


class PerFileConstrainedFileUploadInputFixture(ConstrainedFileUploadInputFixture):
    @scenario
    def size_constrained(self):
        max_size = 13
        self.file_field = FileField(allow_multiple=True, max_size_bytes=max_size, label='Attached files')
        self.validation_error_message = 'files should be smaller than 13.0bytes'
        self.valid_file = temp_file_with('c'*max_size, name='valid_size.html')
        self.invalid_file = temp_file_with('c'*(max_size+1), name='invalid_size.html')

    @scenario
    def type_constrained(self):
        self.file_field = FileField(allow_multiple=True, accept=['text/*'])
        self.validation_error_message = 'files should be of type text/*'
        self.valid_file = temp_file_with('contents', name='valid.html')
        self.invalid_file = temp_file_with('contents', name='invalid.gif')


class MaxNumberOfFilesFileUploadInputFixture(ConstrainedFileUploadInputFixture):
    def new_file_field(self):
        return FileField(allow_multiple=True, max_files=1)


class ToggleableConstraint(ValidationConstraint):
    def __init__(self, fixture=None):
        super(ToggleableConstraint, self).__init__(error_message='test validation message')
        self.fixture = fixture
        
    def validate_input(self, unparsed_input):
        if self.fixture.make_validation_fail:
            raise self

    def __reduce__(self):
        reduced = super(ToggleableConstraint, self).__reduce__()
        pickle_dict = reduced[2]
        del pickle_dict['fixture']
        return reduced


class ToggleValidationFixture(FileUploadInputFixture):
    make_validation_fail = False
    def new_domain_object(self):
        fixture = self
        class DomainObject(object):
            @exposed
            def fields(self, fields):
                fields.files = FileField(allow_multiple=True, label='Attached files')
                fields.files.add_validation_constraint(ToggleableConstraint(fixture))
            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')
                                
        return DomainObject()

    check_script = 'return $(".reahl-nested-form").find(".reload_flag").length > 0'
    def mark_nested_form(self):
        self.driver_browser.execute_script('$(".reahl-nested-form").children().addClass("reload_flag")')
        has_class = self.driver_browser.execute_script(self.check_script)
        assert has_class, 'Something is wrong, could not place flags for checking reloading of form'

    def nested_form_was_reloaded(self):
        has_class = self.driver_browser.execute_script(self.check_script)
        return not has_class  # ie, the UploadPanel has been reloaded


class StubbedFileUploadInputFixture(FileUploadInputFixture):
    run_hook_before = False
    run_hook_after = False
    def new_FileUploadForm(self):
        fixture = self
        class FileUploadInputStub(FileUploadInput):
            def create_html_widget(self):
                return FileUploadPanelStub(self)
            
        class FileUploadPanelStub(FileUploadPanel):
            def upload_file(self):
                if fixture.run_hook_before:
                    fixture.file_upload_hook()
                super(FileUploadPanelStub, self).upload_file()
                if fixture.run_hook_after:
                    fixture.file_upload_hook()

        class FileUploadForm(Form):
            def __init__(self, view):
                super(FileUploadForm, self).__init__(view, 'test')
                self.add_child(FileUploadInputStub(self, fixture.domain_object.fields.files))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))

        return FileUploadForm


class LargeFileUploadInputFixture(StubbedFileUploadInputFixture):
    def file_upload_hook(self):
        self.simulate_large_file_upload()

    def simulate_large_file_upload(self):
        self.upload_done.wait()

    def simulate_large_file_upload_done(self):
        self.upload_done.set()

    @property
    def web_driver(self):  
        return self.run_fixture.chrome_driver  # These tests only work on chrome

    def new_upload_done(self):
        return threading.Event()


class BrokenFileUploadInputFixture(StubbedFileUploadInputFixture):
    run_hook_after = True
    def file_upload_hook(self):
        raise Exception('simulated exception condition')



class FailingConstraint(ValidationConstraint):
    fail = True
    def validate_input(self, unparsed_input):
        if self.fail:
            raise self



@test(FileUploadInputFixture)
def file_upload_input_basics(fixture):
    """A FileUploadInput allows its user to upload multiple files one by one before the Form that
       contains the FileUploadInput is submitted.  When the Form is finally submitted
       the FileField of the FileUploadInput receives all the files uploaded as UploadFile objects.
    """
    fixture.reahl_server.set_app(fixture.wsgi_app)

    browser = fixture.driver_browser
    browser.open('/')

    vassert( not fixture.domain_object.submitted )
    vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
    vassert( not fixture.file_was_uploaded( fixture.file_to_upload2.name ) )

    # Upload one file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))

    vassert( not fixture.domain_object.submitted )
    vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
    vassert( not fixture.file_was_uploaded( fixture.file_to_upload2.name ) )

    # Upload a second file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    vassert( not fixture.domain_object.submitted )
    vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
    vassert( fixture.file_was_uploaded( fixture.file_to_upload2.name ) )

    # Submit the form
    browser.click( XPath.button_labelled('Submit') )
    vassert( fixture.domain_object.submitted )

    # All uploaded files were submitted
    vassert( sorted(fixture.domain_object.submitted_file_info.keys()) == sorted([os.path.basename(f.name) 
                                                                           for f in [fixture.file_to_upload1, fixture.file_to_upload2]] ))

    # Files that were submitted are correct
    file1_content, file1_mime_type = fixture.domain_object.submitted_file_info[os.path.basename(fixture.file_to_upload1.name)]
    vassert( file1_content == fixture.file_to_upload1_content )
    vassert( file1_mime_type == 'text/html' )



@test(FileUploadInputFixture)
def file_upload_input_list_files(fixture):
    """The FileUploadInput displays a list of files that were uploaded so far, but is cleared 
       once the Form is submitted."""
    fixture.reahl_server.set_app(fixture.wsgi_app)

    browser = fixture.driver_browser
    browser.open('/')

    # Upload one file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))

    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    # Upload a second file
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    # Submit the form:
    # If an exception is raised, the list is NOT cleared
    fixture.domain_object.throws_exception = True
    browser.click( XPath.button_labelled('Submit') )
    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    # Upon successful submit, the list IS cleared
    fixture.domain_object.throws_exception = False
    browser.click( XPath.button_labelled('Submit') )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )


@test(FileUploadInputFixture)
def file_upload_input_remove_files(fixture):
    """A user can remove files that were uploaded before the Form which contains the 
       FileUploadInput is submitted."""
    fixture.reahl_server.set_app(fixture.wsgi_app)

    browser = fixture.driver_browser
    browser.open('/')

    # Upload two files
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    # Remove file1
    browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))

    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    # Only the one file is submitted
    browser.click( XPath.button_labelled('Submit') )
    vassert( list(fixture.domain_object.submitted_file_info.keys()) == [fixture.file_to_upload2_name] )


@test(FileUploadInputFixture)
def file_upload_input_double_uploads(fixture):
    """The user is prevented from uploading more than one file with the same name.
    """
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=False))
    browser = fixture.driver_browser
    browser.open('/')

    # Upload two files with the same name
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))

    # Expect an validation error message
    vassert( browser.is_element_present('//span[text()="uploaded files should all have different names"]') )
    vassert( fixture.file_was_uploaded(fixture.file_to_upload1.name) )

@test(FileUploadInputFixture)
def async_upload(fixture):
    """If JavaScript is enabled, the uploading of files happen in the background via ajax (without reloading the page)
       allowing the user to be busy with the rest of the form. The user does not need to click on the Upload button,
       uploading starts automatically upon choosing a file. The list of uploaded files is appropriately updated.
    """
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )

    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    with browser.no_page_load_expected():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)

    vassert( browser.get_value(XPath.input_labelled('Choose file(s)')) == '' )  # Input is cleared for next file to be input
    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )


@test(LargeFileUploadInputFixture)
def async_in_progress(fixture):
    """While a large file is being uploaded, a progress bar and a Cancel button are displayed. Clicking on the Cancel
       button stops the upload and clears the file name from the list of uploaded files.
    """
    fixture.run_hook_before = True
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )

    with fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture

    vassert( browser.is_element_present('//ul/li/progress') )
    progress = browser.get_attribute('//ul/li/progress', 'value')
    vassert( progress == '100' )
    browser.click(XPath.button_labelled('Cancel'))

    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )


@test(LargeFileUploadInputFixture)
def cancelling_queued_upload(fixture):
    """Cancelling an upload that is still queued (upload not started yet) removes the file from the list
       and removed it from the queue of uploads.
    """
    fixture.run_hook_before = True
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    vassert( not fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( not fixture.file_was_uploaded( fixture.file_to_upload2.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    with fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name) # Upload will block, see fixture

    browser.wait_for(fixture.upload_file_is_queued, fixture.file_to_upload1.name)
    browser.wait_for(fixture.upload_file_is_queued, fixture.file_to_upload2.name)

    browser.click(XPath.button_labelled('Cancel', filename=fixture.file_to_upload2_name))

    browser.wait_for_not(fixture.upload_file_is_queued, fixture.file_to_upload2.name)
    fixture.simulate_large_file_upload_done()
    browser.wait_for(fixture.uploaded_file_is_listed, fixture.file_to_upload1.name)

    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( fixture.file_was_uploaded( fixture.file_to_upload1.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )
    vassert( not fixture.file_was_uploaded( fixture.file_to_upload2.name ) )


@test(FileUploadInputFixture)
def prevent_duplicate_upload_js(fixture):
    """The user is prevented from uploading more than one file with the same name on the client side.
    """

    error_locator = XPath.span_containing('uploaded files should all have different names')
    def error_is_visible():
        return browser.is_visible(error_locator)

    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = fixture.driver_browser
    browser.open('/')

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.wait_for_not(error_is_visible)

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.wait_for_not(error_is_visible)

    with fixture.reahl_server.paused():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
        vassert( not fixture.upload_file_is_queued(fixture.file_to_upload1.name) )
        browser.wait_for(error_is_visible)

    browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload2_name))
    browser.wait_for_not(error_is_visible)


@test(LargeFileUploadInputFixture)
def prevent_form_submit(fixture):
    """The user is prevented from submitting the Form while one or more file uploads are still in progress."""
    fixture.run_hook_after = True
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    with fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture

    with browser.no_page_load_expected():
        browser.click( XPath.button_labelled('Submit'), wait=False )

        alert = fixture.web_driver.switch_to.alert
        vassert( alert.text == 'Please try again when all files have finished uploading.' )
        alert.accept()


@test(FileUploadInputFixture)
def async_remove(fixture):
    """With javascript enabled, removing of uploaded files take place via ajax."""
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    # Upload two files
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    browser.click(XPath.button_labelled('Upload'))
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.click(XPath.button_labelled('Upload'))

    # Remove file1
    with browser.no_page_load_expected():
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))

    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    # The javascript works on DOM elements that have been generated server-side as well:
    browser.refresh()
    with browser.no_page_load_expected():
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload2_name))

    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    # No files are submitted eventually
    browser.click( XPath.button_labelled('Submit') )
    vassert( list(fixture.domain_object.submitted_file_info.keys()) == [] )

@test(BrokenFileUploadInputFixture)
def async_upload_error(fixture):
    """If an error happens during (ajax) upload, the user is notified."""
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    fixture.config.reahlsystem.debug = False # So that we don't see the exception output while testing
    browser = fixture.driver_browser
    browser.open('/')

    vassert( not browser.is_element_present(XPath.label_with_text('an error ocurred, please try again later.')) )

    with expected(Exception):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)

    vassert( browser.wait_for_element_present(XPath.span_containing('an error occurred, please try again later.')) )
    vassert( not browser.is_element_enabled(XPath.button_labelled('Cancel')) )

@test(ToggleValidationFixture)
def async_upload_domain_exception(fixture):
    """When a DomainException happens upon uploading via JavaScript, 
       the form is replaced with a rerendered version from the server."""

    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))
    browser = fixture.driver_browser
    browser.open('/')

    fixture.make_validation_fail = False
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)

    fixture.make_validation_fail = True
    fixture.mark_nested_form()
    with browser.no_page_load_expected():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    vassert(fixture.nested_form_was_reloaded())

    # JS Stuff on re-rendered form still work

    # 1: Server-rendered validation message has been cleared
    vassert( browser.is_visible(XPath.span_containing('test validation message')) )
    fixture.make_validation_fail = False
    with browser.no_page_load_expected():
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    browser.wait_for_not(browser.is_visible, XPath.span_containing('test validation message'))

    # 2: The remove button still happens via ajax
    with browser.no_page_load_expected():
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))
        browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload2_name))

@test(LargeFileUploadInputFixture)
def queueing_async_uploads(fixture):
    """Asynchronous uploads do not happen concurrently, they are queued one after another.
    """
    fixture.run_hook_after = True
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    vassert( not fixture.file_was_uploaded(fixture.file_to_upload1.name) )
    vassert( not fixture.uploaded_file_is_listed(fixture.file_to_upload1.name) )

    with fixture.reahl_server.in_background(wait_till_done_serving=False):
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name) # Upload will block, see fixture
        browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name) # Upload will block, see fixture

    progress1 = browser.get_attribute('//ul/li[1]/progress', 'value')
    vassert(progress1 == '100')
    progress2 = browser.get_attribute('//ul/li[2]/progress', 'value')
    vassert(progress2 == '0')

    fixture.simulate_large_file_upload_done()
    browser.wait_for( fixture.uploaded_file_is_listed, fixture.file_to_upload2.name )

    vassert( fixture.uploaded_file_is_listed(fixture.file_to_upload1.name) )
    vassert( fixture.uploaded_file_is_listed(fixture.file_to_upload2.name) )
    vassert( fixture.file_was_uploaded(fixture.file_to_upload1.name) )
    vassert( fixture.file_was_uploaded(fixture.file_to_upload2.name) )

@test(PerFileConstrainedFileUploadInputFixture)
def async_validation(fixture):
    """Validations are checked in JavaScript before uploading.
    """
    # Only tested for the FileUploadInput, as it uses the FileInput
    # in its own implementation, in a NestedForm, and has to pass on the
    # filesize constraint all the way. This way, we test all of that.
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    vassert( not fixture.uploaded_file_is_listed( fixture.valid_file.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.invalid_file.name ) )

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.invalid_file.name)
    vassert( not fixture.uploaded_file_is_listed( fixture.invalid_file.name ) )
    vassert( browser.is_element_present(XPath.span_containing(fixture.validation_error_message)) )

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.valid_file.name)
    vassert( fixture.uploaded_file_is_listed( fixture.valid_file.name ) )

@test(MaxNumberOfFilesFileUploadInputFixture)
def async_number_files_validation(fixture):
    """A Field set to only allow a maximum number of files is checked for validity before uploading in JS.
    """
    # Only tested for the FileUploadInput, as it uses the FileInput
    # in its own implementation, in a NestedForm, and has to pass on the
    # filesize constraint all the way. This way, we test all of that.
    fixture.reahl_server.set_app(fixture.new_wsgi_app(enable_js=True))

    browser = fixture.driver_browser
    browser.open('/')

    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )

    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload1.name)
    vassert( fixture.uploaded_file_is_listed( fixture.file_to_upload1.name ) )
    # Corner case: max are uploaded, but you've not asked to add to them yet:
    vassert( browser.wait_for_not(browser.is_visible, XPath.span_containing('a maximum of 1 files may be uploaded')) )

    # Normal case: max are uploaded, and you're asking to upload another:
    browser.type(XPath.input_labelled('Choose file(s)'), fixture.file_to_upload2.name)
    vassert( not fixture.uploaded_file_is_listed( fixture.file_to_upload2.name ) )
    vassert( browser.wait_for(browser.is_visible, XPath.span_containing('a maximum of 1 files may be uploaded')) )

    browser.click(XPath.button_labelled('Remove', filename=fixture.file_to_upload1_name))
    vassert( browser.wait_for_not(browser.is_visible, XPath.span_containing('a maximum of 1 files may be uploaded')) )



# Copyright 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six
import time


from reahl.tofu import vassert, scenario, expected, test, Fixture
from reahl.stubble import stubclass

from reahl.webdev.tools import XPath
from reahl.web_dev.fixtures import WebFixture, WebBasicsMixin
from reahl.component.exceptions import IsInstance

from reahl.web.fw import Bookmark

from reahl.web.bootstrap.ui import A, Div, P
from reahl.web.bootstrap.forms import Form
from reahl.web.bootstrap.navbar import Navbar, NavbarLayout, ResponsiveLayout
from reahl.web.bootstrap.navs import Nav

from reahl.component.exceptions import ProgrammerError


class NavbarFixture(WebFixture):
    def new_navbar(self):
        return Navbar(self.view)

    def new_navbar_with_layout(self):
        return self.navbar.use_layout(NavbarLayout())

    def new_bookmarks(self):
        return [Bookmark('', '/one', 'One')]

    def new_nav(self):
        return Nav(self.view).with_bookmarks(self.bookmarks)

    def new_form(self):
        return Form(self.view, 'myform')


@test(NavbarFixture)
def navbar_basics(fixture):
    """A typical Navbar is created by using its layout to add some brand text, a nav and form in it."""

    navbar = Navbar(fixture.view).use_layout(NavbarLayout())

    navbar.layout.set_brand_text('Brandy')
    navbar.layout.add(fixture.nav)
    navbar.layout.add(fixture.form)

    [brand, nav, form] = navbar.children[0].children
    [ul] = nav.children

    # The Navbar itself
    vassert( navbar.children[0].tag_name == 'nav' )
    vassert( 'navbar' in navbar.children[0].get_attribute('class').split(' ') )

    # The added contents
    vassert( isinstance(brand, A) )
    vassert( 'navbar-brand' in brand.get_attribute('class') )

    vassert( 'navbar-nav' in ul.get_attribute('class') )

    vassert( isinstance(form, Form))



class LayoutScenarios(NavbarFixture):
    @scenario
    def fixed_top(self):
        self.layout = NavbarLayout(fixed_to='top')
        self.expected_css_class = 'navbar-fixed-top'

    @scenario
    def fixed_bottom(self):
        self.layout = NavbarLayout(fixed_to='bottom')
        self.expected_css_class = 'navbar-fixed-bottom'

    @scenario
    def full(self):
        self.layout = NavbarLayout(full=True)
        self.expected_css_class = 'navbar-full'

    @scenario
    def default(self):
        self.layout = NavbarLayout()
        self.expected_css_class = None


@test(LayoutScenarios)
def navbar_can_have_layout(fixture):
    """NavbarLayout is used to define the placement of a Navbar."""

    widget = Navbar(fixture.view).use_layout(fixture.layout)

    [navbar] = widget.children
    all_classes = ['navbar-fixed-bottom','navbar-fixed-top','navbar-full']
    if fixture.expected_css_class:
        vassert( fixture.expected_css_class in navbar.get_attribute('class').split(' ') )
    
    for not_expected_class in [i for i in all_classes if i != fixture.expected_css_class]:
        vassert( not_expected_class not in navbar.get_attribute('class').split(' ') )


@test(NavbarFixture)
def customised_colour_scheme(fixture):
    """A ColourScheme is used to determine link colours and/or optionally a standard bootstrap background color."""

    layout = NavbarLayout(colour_theme='light', bg_scheme='inverse')
    widget = Navbar(fixture.view).use_layout(layout)

    [navbar] = widget.children

    vassert( 'navbar-light' in navbar.get_attribute('class') )
    vassert( 'bg-inverse' in navbar.get_attribute('class') )


@test(NavbarFixture)
def adding_brand_widget(fixture):
    """Brand content can also be added as a Widget, instead of only text."""

    navbar_widget = fixture.navbar.use_layout(NavbarLayout())
    custom_brand = Div(fixture.view)
    navbar_widget.layout.set_brand(custom_brand)

    [navbar] = navbar_widget.children
    [actual_brand_widget] = navbar.children

    vassert( actual_brand_widget is custom_brand )
    vassert( 'navbar-brand' in actual_brand_widget.get_attribute('class') )


class SpecifyPlacementScenarios(NavbarFixture):
    @scenario
    def adding_left(self):
        self.side = 'left'

    @scenario
    def adding_right(self):
        self.side = 'right'


@test(SpecifyPlacementScenarios)
def adding_to_navbar_with_specific_placement(fixture):
    """Widgets can be added to a Navbar placed right or left in the NavBar."""
    navbar = fixture.navbar.use_layout(NavbarLayout())

    if fixture.side == 'left':
        added_widget = navbar.layout.add(fixture.nav, left=True)
    else:
        added_widget = navbar.layout.add(fixture.nav, right=True)

    [wrapping_div] = navbar.children[0].children

    vassert( 'pull-%s' % fixture.side in wrapping_div.get_attribute('class') )
    vassert( added_widget is fixture.nav )
    vassert( [added_widget] == wrapping_div.children )


@test(SpecifyPlacementScenarios)
def adding_to_navbar_placement_for_device(fixture):
    """Placement of an added Widget can be specified to apply below a certain device size only."""
    fixture.navbar.use_layout(NavbarLayout())

    if fixture.side == 'left':
        fixture.navbar.layout.add(fixture.nav, left='md')
    else:
        fixture.navbar.layout.add(fixture.nav, right='md')
    
    [wrapping_div] = fixture.navbar.children[0].children
    vassert( 'pull-md-%s' % fixture.side in wrapping_div.get_attribute('class') )


@test(NavbarFixture)
def adding_to_navbar_with_both_left_and_right_alignment_not_allowed(fixture):
    """You cannot place an added Widget to both left and right sides."""
    navbar = fixture.navbar.use_layout(NavbarLayout())

    def check_ex(ex):
        vassert( six.text_type(ex).startswith('Both left= and right= have been given. Specify left or right, not both'))

    with expected(ProgrammerError, test=check_ex):
        navbar.layout.add(fixture.nav, left=True, right=True)


@test(NavbarFixture)
def adding_other_than_form_or_nav_is_not_allowed(fixture):
    """Only Navs and Forms may be added."""

    navbar = fixture.navbar.use_layout(NavbarLayout())
    not_a_form_or_nav = Div(fixture.view)

    with expected(IsInstance):
        navbar.layout.add(not_a_form_or_nav)

    # Case: Form
    navbar = fixture.new_navbar().use_layout(NavbarLayout())
    navbar.layout.add(fixture.form)
    [added_widget] = navbar.children[0].children
    vassert( added_widget is fixture.form )

    # Case: Nav
    navbar = fixture.new_navbar().use_layout(NavbarLayout())
    vassert( 'navbar-nav' not in fixture.nav.html_representation.get_attribute('class').split(' ') )
    navbar.layout.add(fixture.nav)
    [added_widget] = navbar.children[0].children
    vassert( added_widget is fixture.nav )
    vassert( 'navbar-nav' in fixture.nav.html_representation.get_attribute('class').split(' ') )


@test(NavbarFixture)
def navbar_with_centered_contents(fixture):
    """Contents of a Navbar appears centered when center_contents is set to True"""

    navbar_widget = fixture.navbar
    navbar_widget.use_layout(NavbarLayout(center_contents=True))
    navbar_widget.layout.set_brand_text('Brandy') #adding something to illustrate the structure change

    [navbar] = navbar_widget.children
    [centering_div] = navbar.children
    [brand_widget] = centering_div.children

    vassert( 'container' in centering_div.get_attribute('class') )
    vassert( 'navbar-brand' in brand_widget.get_attribute('class') )


class NavbarToggleFixture(Fixture, WebBasicsMixin):

    def is_expanded(self, locator):
        return self.driver_browser.is_visible(locator) and \
               self.driver_browser.does_element_have_attribute(locator, 'aria-expanded', value='true') and\
               self.driver_browser.does_element_have_attribute(locator, 'class', value='collapse in') 

    def panel_is_expanded(self):
        return self.is_expanded(XPath.paragraph_containing('Peek-A-Boo'))

    def panel_is_visible(self):
        return self.driver_browser.is_visible(XPath.paragraph_containing('Peek-A-Boo'))

    def xpath_to_locate_toggle(self):
        return XPath('//button[contains(node(), "%s")]' % '☰')

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Div):
            def __init__(self, view):
                super(MainWidget, self).__init__(view)

                navbar = Navbar(view)
                navbar.use_layout(NavbarLayout())
                fixture.element_to_collapse = P(view, text='Peek-A-Boo', css_id='my_id')
                navbar.layout.add_toggle(fixture.element_to_collapse)

                self.add_child(fixture.element_to_collapse)
                self.add_child(navbar)
        return MainWidget


    def new_wsgi_app(self):
        return super(NavbarToggleFixture, self).new_wsgi_app(enable_js=True,
                                                       child_factory=self.MainWidget.factory())
    def new_webconfig(self):
        webconfig = super(NavbarToggleFixture, self).new_webconfig()
        webconfig.frontend_libraries.enable_experimental_bootstrap()
        return webconfig


@test(NavbarToggleFixture)
def navbar_toggle_collapses_html_element(fixture):
    """You can add a toggle to the navbar that hides another element on the page."""

    fixture.reahl_server.set_app(fixture.wsgi_app)
    browser = fixture.driver_browser
    browser.open('/')

    #case: by default, the element to hide is not visible
    vassert( browser.wait_for_not(fixture.panel_is_visible) )

    #case: clicking on the toggle, causes the panel to appear
    browser.click(fixture.xpath_to_locate_toggle())
    browser.wait_for(fixture.panel_is_expanded)

    #case: clicking on the toggle again, causes the panel to disappear
    time.sleep(0.5)
    browser.click(fixture.xpath_to_locate_toggle())
    browser.wait_for_not(fixture.panel_is_visible)


@test(NavbarFixture)
def navbar_toggle_requires_target_id(fixture):
    """To be able to hide an element, it is required to have an id"""

    navbar = fixture.navbar
    navbar.use_layout(NavbarLayout())
    element_without_id = P(fixture.view, text='Peek-A-Boo')

    def check_ex(ex):
        vassert( 'has no css_id set' in six.text_type(ex))

    with expected(ProgrammerError, test=check_ex):
        navbar.layout.add_toggle(element_without_id)


@test(NavbarFixture)
def navbar_toggle_customised(fixture):
    """The text on a toggle that hides an element is customisable"""

    element_to_collapse = Div(fixture.view, css_id='my_id')
    toggle = fixture.navbar_with_layout.layout.add_toggle(element_to_collapse, text='≎')

    [toggle_text_node] = toggle.children

    vassert( '≎' == toggle_text_node.value )


@test(NavbarFixture)
def responsive_navbar(fixture):
    """A ResponsiveLayout hides its Navbar when the viewport becomes smaller than a given device size"""
    navbar_widget = fixture.navbar
    navbar_widget.set_id('my_navbar_id')
    navbar_widget.use_layout(ResponsiveLayout('sm'))

    [navbar] = navbar_widget.children
    [toggle, collapse_div] = navbar.children

    vassert( 'navbar-toggler' in toggle.get_attribute('class') )
    vassert( 'hidden-sm-up' in toggle.get_attribute('class') )
    vassert( 'button' in toggle.get_attribute('type') )
    vassert( 'collapse' in toggle.get_attribute('data-toggle') )
    vassert( 'my_navbar_id' in toggle.get_attribute('data-target') )

    vassert( 'my_navbar_id' in collapse_div.get_attribute('id') )
    vassert( 'collapse' in collapse_div.get_attribute('class') )
    vassert( 'navbar-toggleable-xs' in collapse_div.get_attribute('class') )




# Copyright 2013-2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import re

from webob import Request, Response
from webob.exc import HTTPNotFound

from nose.tools import istest
from reahl.tofu import test, expected, vassert
from reahl.stubble import stubclass, CallMonitor

from reahl.web.fw import Resource, ReahlWSGIApplication, WebExecutionContext, InternalRedirect
from reahl.web.interfaces import UserSessionProtocol
from reahl.web_dev.fixtures import WebFixture, ReahlWSGIApplicationStub
from reahl.webdev.tools import Browser


@istest
class RequestHandlingTests(object):
    class WSGIFixture(WebFixture):
        status = None
        headers = None
        def result_is_valid(self, result):
            return result.startswith('<!DOCTYPE html><html class="no-js">') and result.endswith('</html>')
        def some_headers_are_set(self, headers):
            return dict(headers)['Content-Type'] == 'text/html; charset=utf-8'

    @test(WSGIFixture)
    def wsgi_interface(self, fixture):
        """A ReahlWSGIApplication is a WSGI application."""

        wsgi_app = fixture.wsgi_app

        environ = Request.blank('/', charset='utf8').environ
        def start_response(status, headers):
            fixture.status = status
            fixture.headers = headers
        
        wsgi_iterator = wsgi_app(environ, start_response)

        result = ''.join([i.decode('utf-8') for i in wsgi_iterator])  # To check that it is iterable and get the value
        vassert( fixture.result_is_valid(result) )
        vassert( fixture.status == '200 OK' )
        vassert( fixture.some_headers_are_set(fixture.headers) )

    @test(WebFixture)
    def web_session_handling(self, fixture):
        """The core web framework (this egg) does not implement a notion of session directly.
           It relies on such a notion, but expects an implementation for this to be supplied.

           An implementation should implement the UserSessionProtocol. The implementation
           is thus responsible for associating an instance of itself with the current request.

           The implementation to be used is set in the web.session_class configuration setting.

           The framework ensures that an instance of web.session_class is available during any
           Request. The database is committed before any user code executes, so that any database activity
           done by web.session_class would be committed even if an exception in the user code occurs.
           After user code is executed, methods are called on the web.session_class so that it can set 
           its key on the response and save the last_activity_time.
           Finally, another commit is issued to the database so that any database activity during these last
           actions would also be saved.
        """
        stubclass(UserSessionProtocol)
        class UserSessionStub(UserSessionProtocol):
            session = None
            last_activity_time_set = False
            key_is_set = False
            @classmethod
            def for_current_session(cls):
                assert None, 'Not implemented'

            @classmethod
            def get_or_create_session(cls):
                cls.session = cls()
                return cls.session

            def set_session_key(self, response):
                self.key_is_set = True
                self.saved_response = response

            def is_active(self): pass
            def is_secured(self): pass
            def set_as_logged_in(self, party, stay_logged_in): pass
            def log_out(self): pass

            def set_last_activity_time(self): 
                self.last_activity_time_set = True

            def get_interface_locale(self):
                return 'en_gb'

        # Setting the implementation in config
        fixture.config.web.session_class = UserSessionStub

        with CallMonitor(fixture.system_control.orm_control.commit) as monitor:
            @stubclass(Resource)
            class ResourceStub(object):
                def handle_request(self, request):
                    context = WebExecutionContext.get_context()
                    vassert( context.session is UserSessionStub.session )  # By the time user code executes, the session is set
                    vassert( monitor.times_called == 1 )                      # The database has been committed
                    vassert( not context.session.last_activity_time_set )     
                    vassert( not UserSessionStub.session.key_is_set )
                    return Response()
        
            @stubclass(ReahlWSGIApplication)
            class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
                def resource_for(self, request):
                    return ResourceStub()

            browser = Browser(ReahlWSGIApplicationStub2(fixture.config))

            # A session is obtained, and the correct params passed to the hook methods
            vassert( not UserSessionStub.session )      # Before the request, the session is not yet set
            vassert( monitor.times_called == 0 )           # ... and the database is not yet committed
            browser.open('/')
            
            vassert( monitor.times_called == 1 )           # The database is committed after user code executed
            vassert( UserSessionStub.session )          # The session was set
            vassert( UserSessionStub.session.key_is_set ) # The set_session_key was called
            vassert( UserSessionStub.session.saved_response.status_int is 200 ) # The correct response was passed to set_session_key

            vassert( UserSessionStub.session.last_activity_time_set ) # set_last_activity_time was called

    @test(WebFixture)
    def handling_HTTPError_exceptions(self, fixture):
        """If an HTTPError exception is raised, it is used as response."""
        @stubclass(ReahlWSGIApplication)
        class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
            def resource_for(self, request):
                raise HTTPNotFound()

        browser = Browser(ReahlWSGIApplicationStub2(fixture.config))

        browser.open('/', status=404)

    @test(WebFixture)
    def internal_redirects(self, fixture):
        """During request handling, an InternalRedirect exception can be thrown. This is handled by
           restarting the request loop from scratch to handle the same request again, using a freshly
           constructed resource just as though the request was submitted again by the browser
           save for the browser round trip."""

        fixture.requests_handled = []
        fixture.handling_resources = []

        @stubclass(Resource)
        class ResourceStub(object):
            def handle_request(self, request):
                fixture.requests_handled.append(request)
                fixture.handling_resources.append(self)
                if hasattr(request, 'internal_redirect'):
                    return Response(body='response given after internal redirect')
                raise InternalRedirect(None, None)

        @stubclass(ReahlWSGIApplication)
        class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
            def resource_for(self, request):
                return ResourceStub()

        browser = Browser(ReahlWSGIApplicationStub2(fixture.config))

        browser.open('/')

        vassert( browser.raw_html == 'response given after internal redirect' )
        vassert( fixture.requests_handled[0] is fixture.requests_handled[1] )
        vassert( fixture.handling_resources[0] is not fixture.handling_resources[1] )


    @test(WebFixture)
    def handling_uncaught_exceptions(self, fixture):
        """If an uncaught exception is raised, the session is closed properly."""

        @stubclass(ReahlWSGIApplication)
        class ReahlWSGIApplicationStub2(ReahlWSGIApplicationStub):
            def resource_for(self, request):
                raise AssertionError('this an unknown breakage')
        
        app = ReahlWSGIApplicationStub2(fixture.config)
        browser = Browser(app)

        with CallMonitor(app.system_control.finalise_session) as monitor:
            vassert( monitor.times_called == 0 )
            with expected(AssertionError):
                browser.open('/')
            vassert( monitor.times_called == 1 )




# Copyright 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import vassert, scenario, expected, test

from reahl.web_dev.fixtures import WebFixture

from reahl.web.bootstrap.ui import Alert, HTML5Page



@test(WebFixture)
def alerts(fixture):
    """An alert is used to display a message with a specific severity"""

    alert = Alert(fixture.view, 'Be careful', 'danger')

    vassert( alert.get_attribute('class') == 'alert alert-danger' )
    vassert( alert.get_attribute('role') == 'alert' )
    [message] = alert.children
    vassert( message.value == 'Be careful' )



# Copyright 2015 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import warnings

import six

from reahl.tofu import vassert, scenario, expected, test

from reahl.webdev.tools import XPath, Browser

from reahl.webdev.tools import WidgetTester
from reahl.web_dev.fixtures import WebFixture

from reahl.web.fw import UserInterface
from reahl.web.ui import Div, P, HTML5Page, Header, Footer

from reahl.component.exceptions import ProgrammerError, IsInstance
from reahl.web.pure import ColumnLayout, UnitSize, PageColumnLayout

class ColumnConstructionScenarios(WebFixture):
    @scenario
    def without_sizes(self):
        """Construct the ColumnLayout with a list of column names."""
        self.layout = ColumnLayout('column_a', 'column_b')
        self.expected_class_for_column_b = 'pure-u'

    @scenario
    def with_size(self):
        """You can optionally specify the sizes a column should adhere to."""
        self.layout = ColumnLayout('column_a', ('column_b', UnitSize(default='1/2')))
        self.expected_class_for_column_b = 'pure-u-1-2'


@test(ColumnConstructionScenarios)
def column_layout_basics(fixture):
    """The pure.ColumnLayout adds the correct classes for Pure to lay out its Widget as a pure grid with columns as pure units."""

    widget = Div(fixture.view)
    
    vassert( not widget.has_attribute('class') )
    
    widget.use_layout(fixture.layout)

    vassert( widget.get_attribute('class') == 'pure-g' )
    column_a, column_b = widget.children

    vassert( 'pure-u' in column_a.get_attribute('class') )    # never varies in scenarios
    vassert( fixture.expected_class_for_column_b in column_b.get_attribute('class')  )



class SizingFixture(WebFixture):
    @scenario
    def all_sizes_given(self):
        "The unit_size kwarg can be used to specify sizes for the added column"
        self.sizes = dict(default='1/2', sm='1/3', md='2/3', lg='1/4', xl='3/4')
        self.expected_classes = ['pure-u-1-2','pure-u-lg-1-4','pure-u-md-2-3','pure-u-sm-1-3','pure-u-xl-3-4']

    @scenario
    def no_sizes_specified(self):
        "If no size is given, the column is still a pure unit without a size"
        self.sizes = dict()
        self.expected_classes = ['pure-u']


@test(SizingFixture)
def adding(fixture):
    """You can add a column by calling add_column on the ColumnLayout, only optionally giving a size."""

    widget = Div(fixture.view).use_layout(ColumnLayout())

    widget.layout.add_column(unit_size=UnitSize(**fixture.sizes))

    vassert( widget.children[0].attributes['class'].value == set(fixture.expected_classes) )


@test(WebFixture)
def allowed_sizes(fixture):
    """The device classes for which sizes can be specified, and how to specify a size."""
    size = UnitSize(default='1/2', sm='1/3', md='1/4', lg='1/5', xl='6/7')

    vassert( size == {'default':'1/2', 'sm':'1/3', 'md':'1/4', 'lg':'1/5', 'xl':'6/7'} )


@test(WebFixture)
def page_column_layout_content_layout(fixture):
    """A PageColumnLayout lays out its content (the bits between header and footer) using a ColumnLayout with the columns it is created with."""

    with warnings.catch_warnings(record=True):
        widget = HTML5Page(fixture.view).use_layout(PageColumnLayout('column_a', 'column_b'))
    
    contents_layout = widget.layout.contents.layout
    vassert( isinstance(contents_layout, ColumnLayout) )
    vassert( 'column_a' in contents_layout.columns )
    vassert( 'column_b' in contents_layout.columns )


@test(WebFixture)
def convenient_slots_created(fixture):
    """PageColumnLayout add a Slot for Header, Footer and each column defined. The Slot for each column is named the same as the column's name."""

    with warnings.catch_warnings(record=True):
        layout = PageColumnLayout('column_name_a', 'column_name_b')
    HTML5Page(fixture.view).use_layout(layout)

    header, contents_div, footer = layout.document.children

    vassert( 'header' in header.available_slots )
    vassert( 'footer' in footer.available_slots )

    column_a, column_b = contents_div.children
    vassert( 'column_name_a' in column_a.available_slots )
    vassert( 'column_name_b' in column_b.available_slots )








# Copyright 2013-2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals, absolute_import, division
import six
from reahl.stubble import stubclass
from nose.tools import istest
from reahl.tofu import Fixture
from reahl.tofu import scenario
from reahl.tofu import test
from reahl.tofu import vassert

from reahl.webdev.tools import XPath
from reahl.web.ui import P
from reahl.web.ui import Div
from reahl.web.pager import PageIndex, PagedPanel, PageMenu, SequentialPageIndex, AnnualPageIndex, AnnualItemOrganiserProtocol
from reahl.web_dev.fixtures import WebBasicsMixin


class PageMenuFixture(Fixture, WebBasicsMixin):
    def new_number_of_pages(self):
        return 10
        
    def new_max_page_links(self):
        return 5
        
    def new_PageIndexStub(self):
        fixture = self
        @stubclass(PageIndex)
        class PageIndexStub(PageIndex):
            def __init__(self, max_page_links, number_of_pages):
                super(PageIndexStub, self).__init__(max_page_links=max_page_links)
                self.number_of_pages = number_of_pages

            @property
            def total_number_of_pages(self):
                return self.number_of_pages
            
            def get_contents_for_page(self, page_number):
                return 'contents of page %s' % page_number

            def get_description_for_page(self, page_number):
                return 'p%s' % page_number

        return PageIndexStub
        
    def new_PageContainer(self):
        class PageContainer(PagedPanel):
            def __init__(self, parent, page_index):
                super(PageContainer, self).__init__(parent, page_index, 'container')
                self.add_child(P(self.view, text=self.current_contents))

        return PageContainer

    def new_MainWidget(self):
        fixture = self
        class MainWidget(Div):
            def __init__(self, view):
                super(MainWidget, self).__init__(view)
                page_index = fixture.PageIndexStub(fixture.max_page_links, fixture.number_of_pages)
                page_container = self.add_child(fixture.PageContainer(self.view, page_index))
                self.add_child(PageMenu(self.view, 'page_menu_widget', page_index, page_container))
        return MainWidget
        
    def container_contents_is(self, expected):
        return self.driver_browser.execute_script('return window.jQuery("div p").html() == "%s"' % expected)

    def page_range_links_match(self, link_labels):
        return self.driver_browser.execute_script('return window.jQuery(".reahl-pagemenu a").slice(2,-2).map(function(){return window.jQuery(this).html();}).toArray() == "%s"' % link_labels)

    def new_wsgi_app(self):
        return super(PageMenuFixture, self).new_wsgi_app(enable_js=True, 
                                                       child_factory=self.MainWidget.factory())


@istest
class PageMenuTests(object):
    # Please see fixture for how to declare a PageIndex, PageContainer and PageMenu
    @test(PageMenuFixture)
    def selecting_a_page(self, fixture):
        """Clicking the link of a page results in the contents of the PageContainer being refreshed."""
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open('/')

        fixture.driver_browser.wait_for(fixture.container_contents_is, 'contents of page 1')
        fixture.driver_browser.click(XPath.link_with_text('p2'))
        fixture.driver_browser.wait_for(fixture.container_contents_is, 'contents of page 2')

    @test(PageMenuFixture)
    def navigating_the_page_numbers(self, fixture):
        """One can navigate the range of page links displayed by the PageMenu using the special links."""
        fixture.number_of_pages = 30
        fixture.max_page_links = 5
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open('/')

        # Case: next link
        fixture.driver_browser.click(XPath.link_with_text('>'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, 'p6,p7,p8,p9,p10') )
        fixture.driver_browser.click(XPath.link_with_text('>'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, 'p11,p12,p13,p14,p15') )

        # Case: prev link
        fixture.driver_browser.click(XPath.link_with_text('<'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, 'p6,p7,p8,p9,p10') )

        # Case: last link
        fixture.driver_browser.click(XPath.link_with_text('>|'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, 'p26,p27,p28,p29,p30') )

        # Case: first link
        fixture.driver_browser.click(XPath.link_with_text('|<'))
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, 'p1,p2,p3,p4,p5') )

    @test(PageMenuFixture)
    def contents_when_navigating_the_page_numbers(self, fixture):
        """When navigating the range of page links, the currently displayed contents stay unchanged."""
        fixture.number_of_pages = 30
        fixture.max_page_links = 5
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open('/')

        fixture.driver_browser.click(XPath.link_with_text('p2'))
        fixture.driver_browser.wait_for(fixture.container_contents_is, 'contents of page 2')
        fixture.driver_browser.click(XPath.link_with_text('>'))
        fixture.driver_browser.wait_for(fixture.container_contents_is, 'contents of page 2')

    @test(PageMenuFixture)
    def active_state_of_next_prev_links(self, fixture):
        """Next and Last links are only active when not on the last range of pages,
           and Prev and First are similarly deactive when on the first range of pages."""
        fixture.number_of_pages = 15
        fixture.max_page_links = 5
        fixture.reahl_server.set_app(fixture.wsgi_app)
        fixture.driver_browser.open('/')

        # Case: when you are on the left of the page range        
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text('|<')) )
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text('<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('>')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('>|')) )

        # Case: when you are in the middle of the page range        
        fixture.driver_browser.click(XPath.link_with_text('>'))
        fixture.driver_browser.wait_for_element_present(XPath.link_with_text('p6'))
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('|<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('>')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('>|')) )
        
        # Case: when you are at the end of the page range        
        fixture.driver_browser.click(XPath.link_with_text('>'))
        fixture.driver_browser.wait_for_element_present(XPath.link_with_text('p11'))
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('|<')) )
        vassert( fixture.driver_browser.is_active(XPath.link_with_text('<')) )
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text('>')) )
        vassert( not fixture.driver_browser.is_active(XPath.link_with_text('>|')) )

    class LinkScenarios(PageMenuFixture):
        @scenario
        def only_one_page(self):
            self.number_of_pages = 1
            self.max_page_links = 5
            self.goto_last_range = False
            self.visible_page_descriptions = 'p1'
            self.visible_last_page_descriptions = self.visible_page_descriptions

        @scenario
        def fewer_pages_than_max_links(self):
            self.number_of_pages = 2
            self.max_page_links = 5
            self.goto_last_range = False
            self.visible_page_descriptions = 'p1,p2'
            self.visible_last_page_descriptions = self.visible_page_descriptions

        @scenario
        def more_pages_than_max_links(self):
            self.number_of_pages = 10
            self.max_page_links = 5
            self.goto_last_range = True
            self.visible_page_descriptions = 'p1,p2,p3,p4,p5'
            self.visible_last_page_descriptions = 'p6,p7,p8,p9,p10'

        @scenario
        def more_than_one_range_last_range_not_complete(self):
            self.number_of_pages = 7
            self.max_page_links = 5
            self.goto_last_range = True
            self.visible_page_descriptions = 'p1,p2,p3,p4,p5'
            self.visible_last_page_descriptions = 'p3,p4,p5,p6,p7'

        @scenario
        def changing_number_of_links(self):
            self.number_of_pages = 10
            self.max_page_links = 3
            self.goto_last_range = False
            self.visible_page_descriptions = 'p1,p2,p3'
            self.visible_last_page_descriptions = self.visible_page_descriptions

    @test(LinkScenarios)
    def which_links_display(self, fixture):
        """The menu displays the correct range of page links, depending on the starting page in the range, the
           total number of pages and the max number of links in a range"""
        fixture.reahl_server.set_app(fixture.wsgi_app)

        fixture.driver_browser.open('/')
        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, fixture.visible_page_descriptions) )

        if fixture.goto_last_range:
            fixture.driver_browser.click(XPath.link_with_text('>|'))

        vassert( fixture.driver_browser.wait_for(fixture.page_range_links_match, fixture.visible_last_page_descriptions) )

    @test(Fixture)
    def annual_page_index(self, fixture):
        """The AnnualPageIndex breaks a query of items up into pages containing items with the same year."""
        
        @stubclass(AnnualItemOrganiserProtocol)
        class ItemOrganiser(AnnualItemOrganiserProtocol):
            def get_years(self):                return [2000, 2001]
            def get_items_for_year(self, year): return [year]*3

        page_index = AnnualPageIndex(ItemOrganiser())

        vassert( page_index.get_contents_for_page(2) == [2001, 2001, 2001] )
        vassert( page_index.get_description_for_page(1) == '2000' )
        vassert( page_index.years == [2000, 2001] )
        vassert( page_index.total_number_of_pages == 2 )


    class SequentialScenarios(Fixture):
        def new_items(self):
            return list(range(0, self.number_of_items))

        def new_page_index(self):
            return SequentialPageIndex(self.items, items_per_page=self.items_per_page, max_page_links=12)
            
        @scenario
        def less_than_a_page_items(self):
            self.number_of_items = 2
            self.items_per_page = 10
            self.expected_pages = 1
            self.last_page_contents = self.items[0:2]

        @scenario
        def exactly_one_page_items(self):
            self.number_of_items = 10
            self.items_per_page = 10
            self.expected_pages = 1
            self.last_page_contents = self.items[0:10]

        @scenario
        def just_more_than_one_page_items(self):
            self.number_of_items = 15
            self.items_per_page = 10
            self.expected_pages = 2
            self.last_page_contents = self.items[10:15]
        
    @test(SequentialScenarios)
    def sequential_page_index(self, fixture):
        """The SequentialPageIndex breaks a query of items up into pages of sequential items"""
        
        vassert( fixture.page_index.total_number_of_pages == fixture.expected_pages )
        page = fixture.page_index.get_page_number(fixture.expected_pages)
        vassert( page.description == six.text_type(fixture.expected_pages) )
        vassert( page.contents == fixture.last_page_contents )        
        vassert( fixture.page_index.max_page_links == 12 )





# Copyright 2015, 2016 Reahl Software Services (Pty) Ltd. All rights reserved.
#-*- encoding: utf-8 -*-
#
#    This file is part of Reahl.
#
#    Reahl is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation; version 3 of the License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function, unicode_literals, absolute_import, division

import six

from reahl.tofu import vassert, scenario, expected, test

from reahl.webdev.tools import XPath, Browser

from reahl.webdev.tools import WidgetTester
from reahl.web_dev.fixtures import WebFixture

from reahl.web.fw import UserInterface, Url

from reahl.web_dev.inputandvalidation.inputtests import InputMixin

from reahl.component.exceptions import ProgrammerError, IsInstance
from reahl.component.modelinterface import exposed, Field, BooleanField, Event, Choice, ChoiceField
from reahl.web.bootstrap.ui import A, Div, P, HTML5Page, Header, Footer
from reahl.web.bootstrap.forms import Button, FormLayout, InlineFormLayout, GridFormLayout, Form, ChoicesLayout,\
                                   TextInput, CheckboxInput, PrimitiveCheckboxInput, RadioButtonInput, ButtonLayout
from reahl.web.bootstrap.grid import ResponsiveSize




class FormLayoutScenarios(WebFixture):
    def new_widget(self):
        return Div(self.view).use_layout(self.layout)

    @scenario
    def basic_form(self):
        self.layout = FormLayout()
        self.expected_html = '<div></div>'

    @scenario
    def inline_form(self):
        self.layout = InlineFormLayout()
        self.expected_html = '<div class="form-inline"></div>'

    @scenario
    def grid_form(self):
        self.layout = GridFormLayout(ResponsiveSize(xl=4), ResponsiveSize(xl=8))
        self.expected_html = '<div></div>'


@test(FormLayoutScenarios)
def basic_form_layouts(fixture):
    """There are three basic layouts of forms in bootstrap."""
    tester = WidgetTester(fixture.widget)
    actual = tester.render_html()
    vassert( actual == fixture.expected_html )


class FormLayoutFixture(WebFixture):
    form_group_xpath = '//form/div[contains(@class, "form-group")]'
    def new_domain_object(self):
        class StubDomainObject(object):
            @exposed
            def fields(self, fields):
                fields.an_attribute = Field(label='Some input', required=True)
        return StubDomainObject()
    
    def form_contains_form_group(self, browser):
        return browser.get_xpath_count(self.form_group_xpath) == 1

    def get_form_group_children(self, browser, index=0):
        return browser.xpath( '%s[%s]/*' % (self.form_group_xpath, index+1) )

    def get_form_group(self, browser, index=0):
        form_groups = browser.xpath('%s[%s]' % (self.form_group_xpath, index+1))
        return form_groups[index]

    def get_form_group_highlight_marks(self, browser, index=0):
        form_group = self.get_form_group(browser, index=index)
        return [mark for mark in form_group.attrib['class'].split(' ')
                     if mark.startswith('has-')]
        
    def get_form_group_errors(self, browser, index=0):
        def is_error_element(element):
            return 'class' in element.attrib \
                   and 'text-help' in element.attrib['class'] \
                   and 'has-danger' in element.attrib['class'] 
        def is_visible(element):
            return not (('style' in element.attrib) and ('display: none' in element.attrib['style']))
            
        return [element for element in self.get_form_group_children(browser, index=index)
                        if is_error_element(element) and is_visible(element) ]


@test(FormLayoutFixture)
def adding_basic_input(fixture):
    """Adding an input to a FormLayout, adds it in a bootstrap form-group with Some input."""
    
    class FormWithInputAddedUsingDefaults(Form):
        def __init__(self, view):
            super(FormWithInputAddedUsingDefaults, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute))

    browser = Browser(fixture.new_wsgi_app(child_factory=FormWithInputAddedUsingDefaults.factory()))
    browser.open('/')

    vassert( fixture.form_contains_form_group(browser) )
    
    [label, input_widget] = fixture.get_form_group_children(browser)
    
    # form-group has a label, correctly set up for bootstrap
    vassert( label.tag == 'label' )
    vassert( label.attrib['for'] == input_widget.attrib['id'] )
    vassert( label.text == 'Some input' )
    
    # form-group has an input, correctly set up for bootstrap
    vassert( input_widget.tag == 'input' )
    vassert( input_widget.attrib['name'] == 'an_attribute' )


@test(FormLayoutFixture)
def grid_form_layouts(fixture):
    """A GridFormLayout adds the Label and Input of each added input in separate columns sized like you specify"""
    class FormWithGridFormLayout(Form):
        def __init__(self, view):
            super(FormWithGridFormLayout, self).__init__(view, 'aform')
            self.use_layout(GridFormLayout(ResponsiveSize(lg=4), ResponsiveSize(lg=8)))
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute))

    browser = Browser(fixture.new_wsgi_app(child_factory=FormWithGridFormLayout.factory()))
    browser.open('/')

    [label_column, input_column] = fixture.get_form_group_children(browser)
    vassert( label_column.tag == 'div' )
    vassert( 'col-lg-4' in label_column.attrib['class'] )
    vassert( 'column-label' in label_column.attrib['class'] )

    vassert( input_column.tag == 'div' )
    vassert( 'col-lg-8' in input_column.attrib['class'] )
    vassert( 'column-input' in input_column.attrib['class'] )


@test(FormLayoutFixture)
def specifying_help_text(fixture):
    """You can optionally specify help_text when adding an input."""
    
    class FormWithInputAndHelp(Form):
        def __init__(self, view):
            super(FormWithInputAndHelp, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute), help_text='some help')

    browser = Browser(fixture.new_wsgi_app(child_factory=FormWithInputAndHelp.factory()))
    browser.open('/')

    [label, input_widget, help_text] = fixture.get_form_group_children(browser)

    # form-group has help-text    
    vassert( help_text.tag == 'p' )
    vassert( 'text-muted' in help_text.attrib['class'] )
    vassert( help_text.text == 'some help' )
    

@test(FormLayoutFixture)
def omitting_label(fixture):
    """The label will be rendered hidden (but available to screen readers) if this is explicity requested."""
    class FormWithInputNoLabel(Form):
        def __init__(self, view):
            super(FormWithInputNoLabel, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute), hide_label=True)

    browser = Browser(fixture.new_wsgi_app(child_factory=FormWithInputNoLabel.factory()))
    browser.open('/')

    [label, help_text] = fixture.get_form_group_children(browser)
    vassert( label.tag == 'label' )
    vassert( label.attrib['class'] == 'sr-only' )
    

@test(FormLayoutFixture)
def adding_checkboxes(fixture):
    """CheckboxInputs are added non-inlined, and by default without labels."""

    class DomainObjectWithBoolean(object):
        @exposed
        def fields(self, fields):
            fields.an_attribute = BooleanField(label='Some input', required=True)

    fixture.domain_object = DomainObjectWithBoolean()
    
    class FormWithInputWithCheckbox(Form):
        def __init__(self, view):
            super(FormWithInputWithCheckbox, self).__init__(view, 'aform')
            self.use_layout(FormLayout())
            self.layout.add_input(CheckboxInput(self, fixture.domain_object.fields.an_attribute))

    browser = Browser(fixture.new_wsgi_app(child_factory=FormWithInputWithCheckbox.factory()))
    browser.open('/')

    vassert( not any(child.tag == 'label' for child in fixture.get_form_group_children(browser)) )
    [div] = fixture.get_form_group_children(browser)
    [checkbox] = div.getchildren()
    vassert( checkbox.attrib['class'] == 'checkbox' )
    

class ValidationScenarios(FormLayoutFixture):
    def new_domain_object(self):
        class ModelObject(object):
            @exposed
            def fields(self, fields):
                fields.an_attribute = Field(label='Some input', required=True)
                fields.another_attribute = Field(label='Another input', required=True)
            @exposed
            def events(self, events):
                events.submit = Event(label='Submit')
        return ModelObject()

    def new_Form(self):
        fixture = self
        class FormWithInput(Form):
            def __init__(self, view):
                super(FormWithInput, self).__init__(view, 'aform')
                self.set_attribute('novalidate','novalidate')
                self.use_layout(FormLayout())
                self.layout.add_input(TextInput(self, fixture.domain_object.fields.an_attribute))
                self.layout.add_input(TextInput(self, fixture.domain_object.fields.another_attribute))
                self.define_event_handler(fixture.domain_object.events.submit)
                self.add_child(Button(self, fixture.domain_object.events.submit))
        return FormWithInput

    def new_webconfig(self):
        web = super(ValidationScenarios, self).new_webconfig()
        web.frontend_libraries.enable_experimental_bootstrap()
        return web

    @scenario
    def with_javascript(self):
        self.reahl_server.set_app(self.new_wsgi_app(child_factory=self.Form.factory(), enable_js=True))
        self.browser = self.driver_browser

    @scenario
    def without_javascript(self):
        self.browser = Browser(super(ValidationScenarios, self).new_wsgi_app(child_factory=self.Form.factory()))


@test(ValidationScenarios)
def input_validation_cues(fixture):
    """Visible cues are inserted to indicate the current validation state
       and possible validation error messages to a user. """

    browser = fixture.browser
    browser.open('/')

    vassert( not fixture.get_form_group_highlight_marks(browser, index=0) )
    vassert( not fixture.get_form_group_errors(browser, index=0) )

    browser.type(XPath.input_labelled('Some input'), '')
    browser.click(XPath.button_labelled('Submit'))

    vassert( ['has-danger'] == fixture.get_form_group_highlight_marks(browser, index=0) )
    [error] = fixture.get_form_group_errors(browser, index=0)
    vassert( error.text == 'Some input is required' )

    browser.type(XPath.input_labelled('Some input'), 'valid value')
    browser.click(XPath.button_labelled('Submit'))

    vassert( ['has-success'] == fixture.get_form_group_highlight_marks(browser, index=0) )
    vassert( not fixture.get_form_group_errors(browser, index=0) )

    browser.type(XPath.input_labelled('Another input'), 'valid value')
    browser.click(XPath.button_labelled('Submit'))

    vassert( not fixture.get_form_group_highlight_marks(browser, index=0) )
    vassert( not fixture.get_form_group_errors(browser, index=0) )

@test(ValidationScenarios.with_javascript)
def input_validation_cues_javascript_interaction(fixture):
    """The visual cues rendered server-side can subsequently be manipulated via javascript."""
    fixture.reahl_server.set_app(fixture.new_wsgi_app(child_factory=fixture.Form.factory(), enable_js=False))

    browser = fixture.browser
    browser.open('/')
    browser.type(XPath.input_labelled('Some input'), '')
    browser.click(XPath.button_labelled('Submit'))

    vassert( ['has-danger'] == fixture.get_form_group_highlight_marks(browser, index=0) )
    [error] = fixture.get_form_group_errors(browser, index=0)
    vassert( error.text == 'Some input is required' )

    fixture.reahl_server.set_app(fixture.new_wsgi_app(child_factory=fixture.Form.factory(), enable_js=True))
    browser.open('/')

    browser.click(XPath.button_labelled('Submit'))

    vassert( ['has-danger'] == fixture.get_form_group_highlight_marks(browser, index=0) )
    [error] = fixture.get_form_group_errors(browser, index=0)
    vassert( error.text == 'Some input is required' )

    browser.type(XPath.input_labelled('Some input'), 'valid value')
    browser.press_tab(XPath.input_labelled('Some input'))

    def form_group_is_marked_success(index):
        return ['has-success'] == fixture.get_form_group_highlight_marks(browser, index=index)
    vassert( fixture.driver_browser.wait_for(form_group_is_marked_success, 0) )
    vassert( not fixture.get_form_group_errors(browser, index=0) )


class DisabledScenarios(WebFixture):
    @scenario
    def disabled_input(self):
        self.field = Field(writable=lambda field: False)
        self.expects_disabled_class = True

    @scenario
    def enabled_input(self):
        self.field = Field(writable=lambda field: True)
        self.expects_disabled_class = False


@test(DisabledScenarios)
def disabled_state(fixture):
    """Visible cues are inserted to indicate that inputs are disabled. """
    form = Form(fixture.view, 'test').use_layout(FormLayout())
    field = fixture.field
    field.bind('field', fixture)

    form.layout.add_input(TextInput(form, field))
    
    tester = WidgetTester(form)
    
    [form_group] = tester.xpath(FormLayoutFixture.form_group_xpath)
    if fixture.expects_disabled_class:
        vassert( 'disabled ' in form_group.attrib['class'] )
    else:
        vassert( 'disabled' not in form_group.attrib['class'] )



class ChoicesLayoutFixture(WebFixture):
    def new_form(self):
        return Form(self.view, 'test')

    def new_field(self):
        field = BooleanField()
        field.bind('field', self)
        return field

    def input_is_wrapped_in_label(self, tester):
        return len(tester.xpath('//label/input')) > 0

    def main_element(self, tester):
        return tester.xpath('//div/*')[0]
        

@test(ChoicesLayoutFixture)
def choices_layout(fixture):
    """A ChoicesLayout can be used to add a PrimitiveCheckboxInput inlined or stacked."""
    stacked_container = Div(fixture.view).use_layout(ChoicesLayout())
    stacked_container.layout.add_choice(PrimitiveCheckboxInput(fixture.form, fixture.field))

    tester = WidgetTester(stacked_container)
    vassert( fixture.input_is_wrapped_in_label(tester) )
    vassert( fixture.main_element(tester).tag == 'div' )
    vassert( fixture.main_element(tester).attrib['class'] == 'checkbox' )

    inlined_container = Div(fixture.view).use_layout(ChoicesLayout(inline=True))
    inlined_container.layout.add_choice(PrimitiveCheckboxInput(fixture.form, fixture.field))

    tester = WidgetTester(inlined_container)
    vassert( fixture.input_is_wrapped_in_label(tester) )
    vassert( fixture.main_element(tester).tag == 'label' )
    vassert( fixture.main_element(tester).attrib['class'] == 'checkbox-inline' )


class RadioButtonFixture(ChoicesLayoutFixture):
    def new_field(self):
        choices = [Choice(1, Field(label='One')),
                   Choice(2, Field(label='Two'))
                  ]
        field = ChoiceField(choices)
        field.bind('field', self)
        return field


@test(RadioButtonFixture)
def layout_of_radio_button_input(fixture):
    """The PrimitiveRadioButtonInputs inside a RadioButtonInput are also laid out using a ChoicesLayout."""
    stacked_radio = RadioButtonInput(fixture.form, fixture.field)

    tester = WidgetTester(stacked_radio)
    vassert( fixture.input_is_wrapped_in_label(tester) )
    vassert( fixture.main_element(tester).tag == 'div' )
    vassert( fixture.main_element(tester).attrib['class'] == 'radio' )

    inlined_radio = RadioButtonInput(fixture.form, fixture.field, button_layout=ChoicesLayout(inline=True))

    tester = WidgetTester(inlined_radio)
    vassert( fixture.input_is_wrapped_in_label(tester) )
    vassert( fixture.main_element(tester).tag == 'label' )
    vassert( fixture.main_element(tester).attrib['class'] == 'radio-inline' )



@test(WebFixture)
def button_layouts(fixture):
    """A ButtonLayout can be be used on a Button to customise various visual effects."""

    event = Event(label='click me')
    event.bind('event', fixture)
    form = Form(fixture.view, 'test')
    form.define_event_handler(event)

    # Case: the defaults
    button = Button(form, event).use_layout(ButtonLayout())
    
    tester = WidgetTester(button)
    [button] = tester.xpath(XPath.button_labelled('click me'))
    vassert( button.attrib['class'] == 'btn' )

    # Case: possible effects
    button = Button(form, event).use_layout(ButtonLayout(style='default', size='sm', active=True, wide=True))
    
    tester = WidgetTester(button)
    [button] = tester.xpath(XPath.button_labelled('click me'))
    vassert( button.attrib['class'] == 'active btn btn-block btn-default btn-sm' )


@test(WebFixture)
def button_layouts_on_anchors(fixture):
    """A ButtonLayout can also be used to make an A (anchor) look like a button."""

    anchor = A(fixture.view, href=Url('/an/href'), description='link text').use_layout(ButtonLayout())
    tester = WidgetTester(anchor)
    [rendered_anchor] = tester.xpath(XPath.link_with_text('link text'))
    vassert( rendered_anchor.attrib['class'] == 'btn' )


@test(WebFixture)
def button_layouts_on_disabled_anchors(fixture):
    """Disabled A's are marked with a class so Bootstap can style them appropriately."""
    def can_write():
        return False
    anchor = A(fixture.view, href=Url('/an/href'), description='link text', write_check=can_write).use_layout(ButtonLayout())
    tester = WidgetTester(anchor)
    [rendered_anchor] = tester.xpath(XPath.link_with_text('link text'))
    vassert( rendered_anchor.attrib['class'] == 'btn disabled' )






